import os
from os.path import join, exists
from typing import Tuple, Optional, List, Any

assert List  # make pyflakes happy
assert Any
import hashlib

import click

from dataworkspaces.errors import ConfigurationError, InternalError
from dataworkspaces.workspace import (
    Workspace,
    Resource,
    ResourceRoles,
    SnapshotResourceMixin,
    LocalStateResourceMixin,
    ResourceFactory,
    JSONDict,
)

API_RESOURCE_TYPE = "api-resource"


class ApiResource(Resource, LocalStateResourceMixin, SnapshotResourceMixin):
    """This is a resource type for an API that has to be called to get data.
    It is only valid for source data and intermediate resources.

    To get the hash for a snapshot, a caller needs to go through the data in memory
    and compute the hash. This is usually done by "monkey-patching" a framework's
    API to get the data. The hash is stored in a local scratch directory which
    is read when the snapshot is taken.

    This resource inherits from LocalStateResourceMixin so that we can get
    a clone call to initialze the scratch directory when the workspace or
    individual resource is cloned.
    """

    def __init__(self, name: str, role: str, workspace: Workspace):
        super().__init__(API_RESOURCE_TYPE, name, role, workspace)
        self.hash_states = []  # type: List[Any]

    def validate_subpath_exists(self, subpath: str) -> None:
        raise ConfigurationError(
            "Subpath %s is not valid for resource %s: API resources do not support subpaths"
            % (subpath, self.name)
        )

    def get_local_path_if_any(self) -> Optional[str]:
        return None

    def pull_precheck(self):
        pass

    def pull(self):
        pass

    def push_precheck(self):
        pass

    def push(self):
        pass

    def snapshot_precheck(self) -> None:
        pass

    def snapshot(self) -> Tuple[Optional[str], Optional[str]]:
        scratch = self.workspace._get_local_scratch_space_for_resource(self.name)
        hashfile = join(scratch, "hashval.txt")
        if exists(hashfile):
            with open(hashfile, "r") as f:
                data = f.read().rstrip()
            return (data, None)
        else:
            click.echo("WARNING: no hash available for resource %s" % self.name)
            return (None, None)

    def restore_precheck(self, restore_hashval: str) -> None:
        raise InternalError("Attempt to restore resource %s, which is not restoreable" % self.name)

    def restore(self, restore_hashval: str) -> None:
        raise InternalError("Attempt to restore resource %s, which is not restoreable" % self.name)

    def delete_snapshot(
        self, workspace_snapshot_hash: str, resource_restore_hash: str, relative_path: str
    ) -> None:
        pass

    def init_hash_state(self) -> None:
        """Drop whatever was there before and init with a fresh hash object.
        Use this when starting training.
        """
        self.hash_states = [hashlib.sha1()]

    def dup_hash_state(self) -> None:
        """Push a copy of current TOS. Used when we want to start
        testing. It should already have the training state. We'll
        add a new hash for test data and then pop it when done testing.
        """
        assert len(self.hash_states) > 0
        self.hash_states.append(self.hash_states[-1].copy())

    def pop_hash_state(self) -> None:
        assert len(self.hash_states) > 0
        del self.hash_states[-1]

    def get_hash_state(self):
        assert len(self.hash_states) > 0
        return self.hash_states[-1]

    def save_current_hash(self, comment: Optional[str] = None) -> None:
        """Save the current hash state to the scratch space. If a
        comment is provided, it is written to a separate file.
        """
        assert len(self.hash_states) > 0
        hashval = self.hash_states[-1].hexdigest()
        scratch = self.workspace._get_local_scratch_space_for_resource(self.name)
        hashfile = join(scratch, "hashval.txt")
        with open(hashfile, "w") as f:
            f.write(hashval)
        if self.workspace.verbose:
            print("dws>> %s: wrote hashval of '%s' to %s'" % (self.name, hashval, hashfile))
        commentfile = join(scratch, "comment.txt")
        if comment is not None:
            with open(commentfile, "w") as f:
                f.write(comment + "\n")
        else:
            if exists(commentfile):
                os.remove(commentfile)


class ApiResourceFactory(ResourceFactory):
    def from_command_line(
        self, role: str, name: str, workspace: Workspace, *args, **kwargs
    ) -> Resource:
        """Instantiate a resource object from the add command's
        arguments"""
        if role not in (ResourceRoles.SOURCE_DATA_SET, ResourceRoles.INTERMEDIATE_DATA):
            raise ConfigurationError(
                "API resources only supported for %s and %s roles"
                % (ResourceRoles.SOURCE_DATA_SET, ResourceRoles.INTERMEDIATE_DATA)
            )
        workspace._get_local_scratch_space_for_resource(name, create_if_not_present=True)
        return ApiResource(name, role, workspace)

    def from_json(self, params: JSONDict, local_params: JSONDict, workspace: Workspace) -> Resource:
        """Instantiate a resource object from saved params and local params"""
        return ApiResource(params["name"], params["role"], workspace)

    def has_local_state(self) -> bool:
        """Return true if this resource has local state and needs
        a clone step the first time it is used.

        We return True because we have the local scratch space for the hashes
        that needs to be set up during a clone.
        """
        return True

    def clone(self, params: JSONDict, workspace: Workspace) -> LocalStateResourceMixin:
        """Instantiate a local copy of the resource 
        that came from the remote origin. We don't yet have local params,
        since this resource is not yet on the local machine. If not in batch
        mode, this method can ask the user for any additional information needed
        (e.g. a local path). In batch mode, should either come up with a reasonable
        default or error out if not enough information is available."""
        workspace._get_local_scratch_space_for_resource(params["name"], create_if_not_present=True)
        return ApiResource(params["name"], params["role"], workspace)

    def suggest_name(self, workspace: Workspace, role: str, *args) -> str:
        """Given the arguments passed in to create a resource,
        suggest a name for the case where the user did not provide one
        via --name. This will be used by suggest_resource_name() to
        find a short, but unique name for the resource.
        """
        return role + "-api"



import click

from dataworkspaces.workspace import Workspace, SyncedWorkspaceMixin
from dataworkspaces.errors import ConfigurationError

# XXX remote_repository should be generalized based on the backend type
# and the variants of the publish command we eventually support.
def publish_command(workspace: Workspace, remote_repository: str) -> None:
    if isinstance(workspace, SyncedWorkspaceMixin):
        workspace.publish(remote_repository)
    else:
        raise ConfigurationError(
            "Workspace %s does not support publish command; only supported for synced workspaces"
            % workspace.name
        )

    click.echo("Set remote origin to %s" % remote_repository)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.

import click

from dataworkspaces.workspace import Workspace, SnapshotWorkspaceMixin
from dataworkspaces.errors import ConfigurationError


# XXX port ineage
# def lineage_files_for_snapshot(workspace_dir, snapshot):
#     lineage_dir = get_snapshot_lineage_dir(workspace_dir, snapshot)
#     if isdir(lineage_dir):
#         return {n:join(lineage_dir, n)
#                 for n in os.listdir(lineage_dir) if n.endswith('.json')}
#     else:
#         return {}

# def compare_lineage_files(f1, f2):
#     """Compare two lineage files. Returns a mesasge indicating
#     whether there were differences, and if so, where.
#     """
#     with open(f1, 'r') as f:
#         data1 = json.load(f)
#     with open(f2, 'r') as f:
#         data2 = json.load(f)
#     differences = ''
#     assert data1['step_name']==data2['step_name']
#     for key in ['command_path', 'args', 'cwd', 'timestamp']:
#         v1 = data1[key]
#         v2 = data2[key]
#         if v1==v2:
#             continue
#         differences += "    Values are different for %s:\n" % key
#         differences += "      %s\n" % repr(v1)
#         differences += "      %s\n" % repr(v2)
#     if len(differences)>0:
#         return '  Step %s has lineage differences:\n' % data1['step_name'] + differences
#     else:
#         return '  Step %s has no lineage differences.\n' % data1['step_name']


def diff_command(workspace: Workspace, snapshot_or_tag1: str, snapshot_or_tag2: str) -> None:
    if not isinstance(workspace, SnapshotWorkspaceMixin):
        raise ConfigurationError(
            "Diff command not supported for workspace %s as backend does not support snapshots"
            % workspace.name
        )
    md1 = workspace.get_snapshot_by_tag_or_hash(snapshot_or_tag1)
    manifest1 = {r["name"]: r for r in workspace.get_snapshot_manifest(md1.hashval)}
    snstr1 = "%s, tags %s" % (md1.hashval, ",".join(md1.tags)) if len(md1.tags) > 0 else md1.hashval
    sn1_names = frozenset([n for n in manifest1.keys()])
    md2 = workspace.get_snapshot_by_tag_or_hash(snapshot_or_tag2)
    manifest2 = {r["name"]: r for r in workspace.get_snapshot_manifest(md2.hashval)}
    snstr2 = "%s, tags %s" % (md2.hashval, ",".join(md2.tags)) if len(md2.tags) > 0 else md2.hashval
    sn2_names = frozenset([n for n in manifest2.keys()])

    click.echo("Comparing:\n    Snapshot %s to\n    Snapshot %s" % (snstr1, snstr2))
    common_names = sn1_names.intersection(sn2_names)
    same_resources = []
    different_resources = []
    for name in sorted(common_names):
        h1 = manifest1[name]["hash"]
        h2 = manifest2[name]["hash"]
        if h1 == h2:
            same_resources.append(name)
        else:
            different_resources.append(name)
    if len(same_resources) > 0:
        click.echo("  Resources with the same value:")
        for name in same_resources:
            click.echo("    " + name)
    else:
        click.echo("  Resources with the same value: None")
    if len(different_resources) > 0:
        click.echo("  Resources with different values:")
        for name in different_resources:
            click.echo("    " + name)
    else:
        click.echo("  Resources with different values: None")
    added_resources = sorted(sn2_names.difference(sn1_names))
    if len(added_resources) > 0:
        click.echo("  Added resources:")
        for name in added_resources:
            click.echo("    " + name)
    else:
        click.echo("  Added resources: None")
    removed_resources = sorted(sn1_names.difference(sn2_names))
    if len(removed_resources) > 0:
        click.echo("  Removed resources:")
        for name in removed_resources:
            click.echo("    " + name)
    else:
        click.echo("  Removed resources: None")

    # XXX: port lineage
    # sn1_lineage_files = lineage_files_for_snapshot(workspace_dir, snapshot1)
    # sn1_basenames = set(sn1_lineage_files.keys())
    # sn2_lineage_files = lineage_files_for_snapshot(workspace_dir, snapshot2)
    # sn2_basenames = set(sn2_lineage_files.keys())
    # common_basenames = sn1_basenames.intersection(sn2_basenames)
    # for basename in common_basenames:
    #     print(compare_lineage_files(sn1_lineage_files[basename], sn2_lineage_files[basename]),end='')
    # added_basenames = sn2_basenames.difference(sn1_basenames)
    # if len(added_basenames)>0:
    #     click.echo("  Added lineage steps: %s" % ', '.join([basename.replace('.json', '')
    #                                                         for basename in added_basenames]))
    # else:
    #     click.echo("  No added lineage steps.")
    # removed_basenames = sn1_basenames.difference(sn2_basenames)
    # if len(removed_basenames)>0:
    #     click.echo("  Removed lineage steps: %s" % ', '.join([basename.replace('.json', '')
    #                                                         for basename in removed_basenames]))
    # else:
    #     click.echo("  No removed lineage steps.")



"""Integration with Tensorflow 1.x and 2.0

This is an experimental API and subject to change.

**Wrapping a Karas Model**

Below is an example of wrapping one of the standard tf.keras model classes,
based on https://www.tensorflow.org/tutorials/keras/basic_classification.
Assume we have a workspace already set up, with two resources: a *Source Data*
resource of  type `api-resource`, which is used to capture the hash of
input data as it is passed to the model, and a *Results* resource to
keep the metrics. The only change we need to do to capture the lineage from
the model is to wrap the model's class, using
:func:`~add_lineage_to_keras_model_class`.

Here is the code::
    
    # TensorFlow and tf.keras
    import tensorflow as tf
    from tensorflow import keras
    
    from dataworkspaces.kits.tensorflow1 import add_lineage_to_keras_model_class
    
    # Wrap our model class. This is the only DWS-specific change needed.
    # We add an optional checkpoint configuration, which will cause checkpoints
    # to be written to the workspace's scratch directory and then the best
    # checkpoint copied to the results resource.
    keras.Sequential = add_lineage_to_keras_model_class(keras.Sequential,
                           checkpoint_config=CheckpointConfig(model='fashion',
                                                              monitor='loss'))
    
    fashion_mnist = keras.datasets.fashion_mnist
    
    (train_images, train_labels), (test_images, test_labels) = fashion_mnist.load_data()
    
    model = keras.Sequential([
        keras.layers.Flatten(input_shape=(28, 28)),
        keras.layers.Dense(128, activation=tf.nn.relu),
        keras.layers.Dense(10, activation=tf.nn.softmax)
    ])
    
    model.compile(optimizer='adam',
                  loss='sparse_categorical_crossentropy',
                  metrics=['accuracy'])
    
    model.fit(train_images, train_labels, epochs=5)
    
    test_loss, test_acc = model.evaluate(test_images, test_labels)
    print('Test accuracy:', test_acc)

This will create a ``results.json`` file in the results resource. It will
look like this::

    {
      "step": "test",
      "start_time": "2019-09-26T11:33:22.100584",
      "execution_time_seconds": 26.991521,
      "parameters": {
        "optimizer": "adam",
        "loss_function": "sparse_categorical_crossentropy",
        "epochs": 5,
        "fit_batch_size": null,
        "evaluate_batch_size": null
      },
      "run_description": null,
      "metrics": {
        "loss": 0.3657455060243607,
        "acc": 0.8727999925613403
      }
    }


**Subclassing from a Keras Model**

If you subclass from a Keras Model class, you can just use
:func:`~add_lineage_to-keras_model_class` as a decorator. Here is an example::
    
    @add_lineage_to_keras_model_class
    class MyModel(keras.Model):
      def __init__(self):
        # The Tensorflow documentation tends to specify the class name
        # when calling the superclass __init__ function. Don't do this --
        # it breaks if you use class decorators!
        #super(MyModel, self).__init__()
        super().__init__()
        self.dense1 = tf.keras.layers.Dense(4, activation=tf.nn.relu)
        self.dense2 = tf.keras.layers.Dense(5, activation=tf.nn.softmax)
    
      def call(self, inputs):
        x1 = self.dense1(inputs)
        return self.dense2(x1)
    
    model = MyModel()
    
    import numpy as np
    
    model.compile(optimizer='adam',
                  loss='sparse_categorical_crossentropy',
                  metrics=['accuracy'])
    model.fit(np.zeros(20).reshape((5,4)), np.ones(5), epochs=5)
    test_loss, test_acc = model.evaluate(np.zeros(16).reshape(4,4), np.ones(4))
    
    print('Test accuracy:', test_acc)


**Supported datatypes for API Resources**

If you are using the *API Resource Type* for your input resource,
the model wrapper will hash the incoming data parameters and include
the hash values in the data lineage. To compute the hashes, Data
Workspaces must access the underlying data representation. The following data
types are currently supported:

* NumPy ``ndarray``
* Pandas ``DataFrame`` and ``Series``
* Tensorflow ``Tensor`` and ``Dataset``, as well as tuples and dictionaries
  containing these types. These types supported if you are either running
  Tensorflow 2.x (graph or eager mode) or 1.x only in eager mode. This restriction is
  due to the inability to access the underlying tensor representation when Tensorflow
  is running in graph mode in version 1.x

If you are using another data representation, or running Tensorflow 1.x in graph
mode, you can always use a resource type that stores the data in files
(e.g. git or local-files) and pass in the input resource name to the
wrapper function.


**API**

"""
from typing import Optional, Union, List, Dict, cast, NamedTuple

assert List
import os
from os.path import join, isdir, exists, basename
import re
import glob
from types import GeneratorType

import tensorflow

if tensorflow.__version__.startswith("2."):  # type: ignore
    USING_TENSORFLOW2 = True
else:
    USING_TENSORFLOW2 = False
import tensorflow.keras.optimizers as optimizers
import tensorflow.keras.utils as kerasutils
from tensorflow.keras.callbacks import ModelCheckpoint

if USING_TENSORFLOW2:
    import tensorflow.keras.losses as losses
else:
    import tensorflow.losses as losses

from dataworkspaces.workspace import (
    find_and_load_workspace,
    ResourceRef,
    ResourceRoles,
    FileResourceMixin,
)
from dataworkspaces.errors import ConfigurationError
from dataworkspaces.kits.wrapper_utils import (
    _DwsModelState,
    _add_to_hash,
    NotSupportedError,
    _find_resource,
)


def _verify_eager_if_dataset(x, y, api_resource):
    """If this is tensorflow 1.x and non-eager mode, there's no way
    to evaluate the dataset outside the tensor graph.
    """
    if (
        (not USING_TENSORFLOW2)
        and (isinstance(x, tensorflow.data.Dataset) or isinstance(y, tensorflow.data.Dataset))  # type: ignore
        and (not tensorflow.executing_eagerly())  # type: ignore
    ):
        raise NotSupportedError(
            "Using an API resource ("
            + api_resource.name
            + ") with non-eager datasets is not "
            + "supported with TensorFlow 1.x."
        )


def _wrap_generator(wrapped, hash_state):
    """Return a generator such that it hashes
    the values returned for each iterator
    """

    def wrapper():
        for v in wrapped:
            if len(v) == 2:
                (inputs, targets) = v
                sample_weights = None
            else:
                (inputs, targets, sample_weights) = v
            _add_to_hash(inputs, hash_state)
            _add_to_hash(targets, hash_state)
            if sample_weights is not None:
                _add_to_hash(sample_weights, hash_state)
            yield v

    return wrapper()


class _TfKerasSequenceWrapper(kerasutils.Sequence):
    def __init__(self, wrapped, hash_state):
        self.wrapped = wrapped
        self.hash_state = hash_state

    def __getitem__(self, idx):
        v = self.wrapped.__getitem__(idx)
        if len(v) == 2:
            (inputs, targets) = v
            sample_weights = None
        else:
            (inputs, targets, sample_weights) = v
        _add_to_hash(inputs, self.hash_state)
        _add_to_hash(targets, self.hash_state)
        if sample_weights is not None:
            _add_to_hash(sample_weights, self.hash_state)
        return v

    def __len__(self):
        return self.wrapped.__len__()

    def __iter__(self):
        return _wrap_generator(self.wrapped, self.hash_state)

    def on_epoch_end(self):
        return self.wrapped.on_epoch_end()


class DwsModelCheckpoint(ModelCheckpoint):
    """
    Subclass of tf.keras.callbacks.ModelCheckpoint which will save checkpoints
    to the workspace's stratch space and then move the most recent/best checkpoint
    to the results directory at the end of the run.

    You can instantiate this class directly and pass it to the ``callbacks``
    parameter of the model's ``fit()`` method::

          model.fit(train_images, train_labels, epochs=10,
                    callbacks=[DwsModelCheckpoint('fashion', monitor='loss', save_best_only=True)])

    You can also pass :class:`~CheckpointConfig` instance to the
    :func:`~add_lineage_to_keras_model_class` wrapper function.
    """

    def __init__(
        self,
        model_name: str,
        monitor: str = "val_loss",
        save_best_only: bool = False,
        mode: str = "auto",
        save_freq: Union[str, int] = "epoch",
        results_resource: Optional[Union[str, ResourceRef]] = None,
        workspace_dir: Optional[str] = None,
        verbose: Union[int, bool] = 0,
    ):
        """
        model_name is used to create the checkpoint filenames. The checkpoints
        will be saved as MODEL_NAME_{epoch}.

        Currently, only supports save_weights_only option.

        verbose can be either 0,1 in the style of tensorflow or a True,False
        in the style of Data Workspaces.

        """
        self.dws_model_name = model_name
        if verbose == 0 or verbose == False:
            tf_verbose = 0
            dws_verbose = False
        else:
            tf_verbose = 1
            dws_verbose = True

        self.workspace = find_and_load_workspace(
            batch=True, verbose=dws_verbose, uri_or_local_path=workspace_dir
        )

        results_ref = _find_resource(self.workspace, ResourceRoles.RESULTS, results_resource)
        self.results_resource = self.workspace.get_resource(results_ref.name)
        if not isinstance(self.results_resource, FileResourceMixin):
            raise ConfigurationError("Resource %s is not a file-based resource" % results_ref.name)
        self.results_subdir = results_ref.subpath  # type: Optional[str]
        scratch_dir = self.workspace.get_scratch_directory()
        assert isdir(scratch_dir), "missing scratch directory %s" % scratch_dir
        self.dws_checkpoint_path = join(scratch_dir, "checkpoints")  # type: str
        if not isdir(self.dws_checkpoint_path):
            os.mkdir(self.dws_checkpoint_path)
        self.checkpoint_filepath_template = join(self.dws_checkpoint_path, model_name + "_{epoch}")
        super().__init__(
            filepath=self.checkpoint_filepath_template,
            monitor=monitor,
            save_best_only=save_best_only,
            mode=mode,
            save_freq=save_freq,
            save_weights_only=True,
            verbose=tf_verbose,
        )

    def on_train_begin(self, logs: Optional[Dict] = None):
        files_to_delete = []  # type: List[str]
        files_to_delete.extend(
            glob.glob(join(self.dws_checkpoint_path, self.dws_model_name + "_*[0-9].index"))
        )
        files_to_delete.extend(
            glob.glob(
                join(
                    self.dws_checkpoint_path, self.dws_model_name + "_*[0-9].data-*[0-9]-of-*[0-9]"
                )
            )
        )
        checkpoint_metadata_file = join(self.dws_checkpoint_path, "checkpoint")
        if exists(checkpoint_metadata_file):
            files_to_delete.append(checkpoint_metadata_file)
        for f in files_to_delete:
            os.remove(f)
        print(
            "dws> Removed %d old checkpoint files for model %s ahead of training"
            % (len(files_to_delete), self.dws_model_name)
        )
        return super().on_train_begin(logs)

    def on_train_end(self, logs: Optional[Dict] = None):
        checkpoint_metadata_file = join(self.dws_checkpoint_path, "checkpoint")
        assert exists(checkpoint_metadata_file), (
            "Missing checkpoint metadata file %s" % checkpoint_metadata_file
        )
        # find the checkpoint that we want to save
        with open(checkpoint_metadata_file, "r") as f:
            MODEL_CHECKPOINT_PATH = re.compile(
                "^"
                + re.escape("model_checkpoint_path:")
                + r'\s+"('
                + re.escape(self.dws_model_name + "_")
                + r'\d+)"$'
            )
            checkpoint_base = None
            for line in f:
                mo = MODEL_CHECKPOINT_PATH.match(line.rstrip())
                if mo is not None:
                    checkpoint_base = mo.group(1)
                    break
            assert checkpoint_base is not None, (
                "Did not find model checkpoint path in %s" % checkpoint_metadata_file
            )
        copy_files = []  # type: List[str]
        copy_files.append(join(self.dws_checkpoint_path, checkpoint_base + ".index"))
        copy_files.extend(
            glob.glob(join(self.dws_checkpoint_path, checkpoint_base + ".data-*[0-9]-of-*[0-9]"))
        )
        copy_files.append(
            join(self.dws_checkpoint_path, "checkpoint")
        )  # copy index file to make it easy to load checkpoint
        for src_file in copy_files:
            if self.results_subdir is not None:
                dest_path = join(self.results_subdir, basename(src_file))
            else:
                dest_path = basename(src_file)
            cast(FileResourceMixin, self.results_resource).upload_file(src_file, dest_path)
        if self.results_subdir is not None:
            print(
                "dws> Copied checkpoint %s to resource %s:%s"
                % (checkpoint_base, self.results_resource.name, self.results_subdir)
            )
        else:
            print(
                "dws> Copied checkpoint %s to resource %s"
                % (checkpoint_base, self.results_resource.name)
            )

        return super().on_train_end(logs)


class CheckpointConfig(NamedTuple):
    """Configuration for checkpoints, to be passed as a parameter
    to :func:`~add_lineage_to_keras_model_class`, instead of
    directly instantiating :class:`~DwsModelChecpoint`.

    The checkpoints are initially written under the workspace's
    scratch space. At the end of training, the best checkpoint is
    copied to the results resource.

    The configuration fields are:

    * ``model_name`` - name of the model to use in checkpoint files
    * ``monitor`` - metric to monitor - defaults to val_loss
    * ``save_best_only`` - if True, only checkpoints better than the
      previous are kept.
    * ``mode`` - how to determine whether a metric is the "best" - auto, min, or max
    * ``save_freq`` - 'epoch' or an interger
    """

    model_name: str
    monitor: str = "val_loss"
    save_best_only: bool = False
    mode: str = "auto"
    save_freq: Union[str, int] = "epoch"


def add_lineage_to_keras_model_class(
    Cls: type,
    input_resource: Optional[Union[str, ResourceRef]] = None,
    results_resource: Optional[Union[str, ResourceRef]] = None,
    workspace_dir: Optional[str] = None,
    checkpoint_config: Optional[CheckpointConfig] = None,
    verbose: bool = False,
) -> type:
    """This function wraps a Keras model class with a subclass that overwrites
    key methods to make calls to the data lineage API.

    **Parameters:**

    * ``Cls`` -- the class being wrapped
    * ``input_resources`` -- optional list of input resources to this model.
      Each resource may be specified by name, by a local file path, or via a
      ``ResourceRef``. If no inputs are specified, will try to infer from the
      workspace.
    * ``results_resource`` -- optional resource where the results are to be stored.
      May be specified by name, by a local file path, or via a ``ResourceRef``.
      if not specified, will try to infer from the workspace.
    * ``workspace-dir`` -- Optional directory specifying the workspace. Usually can be
      inferred from the current directory.
    * ``checkpoint_config`` -- Optional instance of :class:`~CheckpointConfig`, which
      is used to enable checkpointing on fit and fit_generator()
    * ``verbose`` -- If True, print extra debugging information.

    The following methods are wrapped:

    * :func:`~__init__` - loads the workspace and adds dws-specific class members
    * :func:`~compile` - captures the ``optimizer`` and ``loss_function`` parameter values
    * :func:`~fit` - captures the ``epochs`` and ``batch_size`` parameter values;
      if input is an API resource, capture hash values of training data, otherwise capture
      input resource name. If the input is an API resource, and it is either a Keras Sequence
      or a generator, writes the generator and captures the hashes of returned values as it
      is iterated through.
    * :func:`~evaluate` - captures the ``batch_size`` parameter value; if input is an
      API resource, capture hash values of test data, otherwise capture input resource
      name; capture metrics and write them to results resource. If the input is an API resource,
      and it is either a Keras Sequence or a generator, writes the generator and captures
      the hashes of returned values as it is iterated through.
    """
    if hasattr(Cls, "_dws_model_wrap") and Cls._dws_model_wrap is True:  # type: ignore
        print("dws>> %s or a superclass is already wrapped" % Cls.__name__)
        return Cls  # already wrapped
    workspace = find_and_load_workspace(
        batch=True, verbose=verbose, uri_or_local_path=workspace_dir
    )

    class WrappedModel(Cls):  # type: ignore
        _dws_model_wrap = True

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._dws_state = _DwsModelState(workspace, input_resource, results_resource)
            if checkpoint_config is not None:
                self.checkpoint_cb = DwsModelCheckpoint(
                    checkpoint_config.model_name,
                    monitor=checkpoint_config.monitor,
                    save_best_only=checkpoint_config.save_best_only,
                    mode=checkpoint_config.mode,
                    save_freq=checkpoint_config.save_freq,
                    results_resource=results_resource,
                    workspace_dir=workspace_dir,
                    verbose=verbose,
                )  # type: Optional[DwsModelCheckpoint]
            else:
                self.checkpoint_cb = None

        def compile(
            self,
            optimizer,
            loss=None,
            metrics=None,
            loss_weights=None,
            sample_weight_mode=None,
            weighted_metrics=None,
            target_tensors=None,
            distribute=None,
            **kwargs,
        ):
            if isinstance(optimizer, str):
                self._dws_state.lineage.add_param("optimizer", optimizer)
            elif isinstance(optimizer, optimizers.Optimizer):
                self._dws_state.lineage.add_param("optimizer", optimizer.__class__.__name__)
            if isinstance(loss, str):
                self._dws_state.lineage.add_param("loss_function", loss)
            elif isinstance(loss, losses.Loss):
                self._dws_state.lineage.add_param("loss_function", loss.__class__.__name__)
            if tensorflow.__version__<"2.2.": # type: ignore
                return super().compile(
                    optimizer,
                    loss,
                    metrics,
                    loss_weights,
                    sample_weight_mode,
                    weighted_metrics,
                    target_tensors,
                    distribute,
                    **kwargs,
                )
            else: # starting in 2.2, tensorflow removed the tartet_tensors and distribute args
                return super().compile(
                    optimizer,
                    loss,
                    metrics,
                    loss_weights,
                    sample_weight_mode,
                    weighted_metrics,
                    **kwargs,
                )

        def fit(self, x, y=None, **kwargs):
            """x, y can be arrays or x can be a generator.
            """
            if "epochs" in kwargs:
                self._dws_state.lineage.add_param("fit.epochs", kwargs["epochs"])
            else:
                self._dws_state.lineage.add_param("fit.epochs", 1)
            if "batch_size" in kwargs:
                self._dws_state.lineage.add_param("fit.batch_size", kwargs["batch_size"])
            else:
                self._dws_state.lineage.add_param("fit.batch_size", None)
            api_resource = self._dws_state.find_input_resources_and_return_if_api(x, y)
            if api_resource is not None:
                _verify_eager_if_dataset(x, y, api_resource)
                api_resource.init_hash_state()
                hash_state = api_resource.get_hash_state()
                if isinstance(x, kerasutils.Sequence):
                    if y is not None:
                        raise NotSupportedError(
                            "fit() method does not suppport a generator for x AND a y value"
                        )
                    x = _TfKerasSequenceWrapper(x, hash_state)
                elif isinstance(x, GeneratorType):
                    if y is not None:
                        raise NotSupportedError(
                            "fit() method does not suppport a generator for x AND a y value"
                        )
                    x = _wrap_generator(x, hash_state)
                else:  # x and y are provided as full arrays
                    _add_to_hash(x, hash_state)
                    if y is not None:
                        _add_to_hash(y, hash_state)
                    api_resource.save_current_hash()  # in case we evaluate in a separate process
            if self.checkpoint_cb:
                if "callbacks" in kwargs:
                    kwargs["callbacks"].append(self.checkpoint_cb)
                else:
                    kwargs["callbacks"] = [
                        self.checkpoint_cb,
                    ]
            return super().fit(x, y, **kwargs)

        def fit_generator(
            self,
            generator,
            steps_per_epoch=None,
            epochs=1,
            verbose=1,
            callbacks=None,
            validation_data=None,
            validation_steps=None,
            validation_freq=1,
            class_weight=None,
            max_queue_size=10,
            workers=1,
            use_multiprocessing=False,
            shuffle=True,
            initial_epoch=0,
        ):
            self._dws_state.lineage.add_param("fit_generator.epochs", epochs)
            self._dws_state.lineage.add_param("fit_generator.steps_per_epoch", steps_per_epoch)
            api_resource = self._dws_state.find_input_resources_and_return_if_api(generator)
            if api_resource is not None:
                # wrap the generator to capture each entry as it is returned
                api_resource.init_hash_state()
                hash_state = api_resource.get_hash_state()
                if isinstance(generator, kerasutils.Sequence):
                    generator = _TfKerasSequenceWrapper(generator, hash_state)
                else:
                    generator = _wrap_generator(generator, hash_state)
            if self.checkpoint_cb:
                if callbacks is not None:
                    callbacks.append(self.checkpoint_cb)
                else:
                    callbacks = [
                        self.checkpoint_cb,
                    ]
            results = super().fit_generator(
                generator,
                steps_per_epoch,
                epochs,
                verbose,
                callbacks,
                validation_data,
                validation_steps,
                validation_freq,
                class_weight,
                max_queue_size,
                workers,
                use_multiprocessing,
                shuffle,
                initial_epoch,
            )
            if api_resource is not None:
                api_resource.save_current_hash()
            return results

        def evaluate(self, x, y=None, **kwargs):
            if "batch_size" in kwargs:
                self._dws_state.lineage.add_param("evaluate.batch_size", kwargs["batch_size"])
            else:
                self._dws_state.lineage.add_param("evaluate.batch_size", None)
            api_resource = self._dws_state.find_input_resources_and_return_if_api(x, y)
            if api_resource is not None:
                _verify_eager_if_dataset(x, y, api_resource)
                api_resource.dup_hash_state()
                hash_state = api_resource.get_hash_state()
                if isinstance(x, kerasutils.Sequence):
                    if y is not None:
                        raise NotSupportedError(
                            "evaluate() method does not suppport a generator for x AND a y value"
                        )
                    x = _TfKerasSequenceWrapper(x, hash_state)
                elif isinstance(x, GeneratorType):
                    if y is not None:
                        raise NotSupportedError(
                            "evaluate() method does not suppport a generator for x AND a y value"
                        )
                    x = _wrap_generator(x, hash_state)
                else:
                    _add_to_hash(x, hash_state)
                    if y is not None:
                        _add_to_hash(y, hash_state)
            results = super().evaluate(x, y, **kwargs)
            assert len(results) == len(self.metrics_names)
            if api_resource is not None:
                api_resource.save_current_hash()
                api_resource.pop_hash_state()
            self._dws_state.write_metrics_and_complete(
                {n: v for (n, v) in zip(self.metrics_names, results)}
            )
            return results

        def evaluate_generator(
            self,
            generator,
            steps=None,
            callbacks=None,
            max_queue_size=10,
            workers=1,
            use_multiprocessing=False,
            verbose=0,
        ):
            self._dws_state.lineage.add_param("evaluate_generator.steps", steps)
            api_resource = self._dws_state.find_input_resources_and_return_if_api(generator)
            if api_resource is not None:
                # wrap the generator to capture each entry as it is returned
                api_resource.dup_hash_state()
                hash_state = api_resource.get_hash_state()
                if isinstance(generator, kerasutils.Sequence):
                    generator = _TfKerasSequenceWrapper(generator, hash_state)
                else:
                    generator = _wrap_generator(generator, hash_state)
            results = super().evaluate_generator(
                generator, steps, callbacks, max_queue_size, workers, use_multiprocessing, verbose
            )
            if api_resource is not None:
                api_resource.save_current_hash()
                api_resource.pop_hash_state()
            assert len(results) == len(self.metrics_names)
            self._dws_state.write_metrics_and_complete(
                {n: v for (n, v) in zip(self.metrics_names, results)}
            )
            return results

    WrappedModel.__name__ = Cls.__name__  # this is to fake things out for the reporting
    if workspace.verbose:
        print("dws>> Wrapped model class %s" % Cls.__name__)
    return WrappedModel



#!/usr/bin/env python3
# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Command-line tool for data workspaces
"""

__all__ = ["cli"]
import sys
import click
import re
from os.path import isdir, join, abspath, expanduser, basename, curdir
from typing import Optional, Union
from argparse import Namespace
from collections.abc import Sequence

from dataworkspaces import __version__
from dataworkspaces.commands.init import init_command
from dataworkspaces.commands.add import add_command
from dataworkspaces.commands.snapshot import snapshot_command
from dataworkspaces.commands.delete_snapshot import delete_snapshot_command
from dataworkspaces.commands.restore import restore_command
from dataworkspaces.commands.status import status_command
from dataworkspaces.commands.report import (
    report_status_command,
    report_history_command,
    report_lineage_command,
    report_results_command,
)
from dataworkspaces.commands.publish import publish_command
from dataworkspaces.commands.push import push_command
from dataworkspaces.commands.pull import pull_command
from dataworkspaces.commands.clone import clone_command

# from dataworkspaces.commands.run import run_command
from dataworkspaces.commands.diff import diff_command
from dataworkspaces.commands.lineage import lineage_graph_command
from dataworkspaces.commands.deploy import deploy_build_command, deploy_run_command
from dataworkspaces.commands.config import config_command
from dataworkspaces.workspace import (
    RESOURCE_ROLE_CHOICES,
    ResourceRoles,
    find_and_load_workspace,
    _find_containing_workspace,
)
from dataworkspaces.errors import BatchModeError
from dataworkspaces.utils.param_utils import DEFAULT_HOSTNAME
from dataworkspaces.utils.regexp_utils import HOSTNAME_RE
from dataworkspaces.utils.file_utils import LocalPathType

CURR_DIR = abspath(expanduser(curdir))
CURR_DIRNAME = basename(CURR_DIR)
DWS_PATHDIR = _find_containing_workspace()

# we are going to store the verbose mode
# in a global here and wrap it in a function
# so that we can access it from __main__.
VERBOSE_MODE = False


def is_verbose_mode():
    global VERBOSE_MODE
    return VERBOSE_MODE


class WorkspaceDirParamType(click.ParamType):
    name = "workspace-directory"

    def convert(self, value, param, ctx):
        path = abspath(expanduser(value))
        wspath = join(path, ".dataworkspace")
        if not isdir(path):
            self.fail("Directory '%s' does not exist" % value, param, ctx)
        elif not isdir(wspath):
            self.fail(
                "No .dataworkspace directory found under '%s'. Did you run 'dws init'?" % value,
                param,
                ctx,
            )
        else:
            return path


WORKSPACE_PARAM = WorkspaceDirParamType()


class DirectoryParamType(click.ParamType):
    name = "directory"

    def convert(self, value, param, ctx):
        path = abspath(expanduser(value))
        if not isdir(path):
            self.fail("Directory '%s' does not exist" % value, param, ctx)
        else:
            return path


DIRECTORY_PARAM = DirectoryParamType()


class RoleParamType(click.ParamType):
    name = "role (one of %s)" % ", ".join(RESOURCE_ROLE_CHOICES)

    def convert(self, value, param, ctx):
        value = value.lower()
        if value in (ResourceRoles.SOURCE_DATA_SET, "s"):
            return ResourceRoles.SOURCE_DATA_SET
        elif value in (ResourceRoles.INTERMEDIATE_DATA, "i"):
            return ResourceRoles.INTERMEDIATE_DATA
        elif value in (ResourceRoles.CODE, "c"):
            return ResourceRoles.CODE
        elif value in (ResourceRoles.RESULTS, "r"):
            return ResourceRoles.RESULTS
        else:
            self.fail(
                "Invalid resource role. Must be one of: %s" % ", ".join(RESOURCE_ROLE_CHOICES)
            )


ROLE_PARAM = RoleParamType()
ROLE_DATA_CHOICES = [ResourceRoles.SOURCE_DATA_SET, ResourceRoles.INTERMEDIATE_DATA]


class DataRoleParamType(click.ParamType):
    """A role parameter limited to source and intermediate data."""

    name = "role (one of %s)" % ", ".join(ROLE_DATA_CHOICES)

    def convert(self, value, param, ctx):
        value = value.lower()
        if value in (ResourceRoles.SOURCE_DATA_SET, "s"):
            return ResourceRoles.SOURCE_DATA_SET
        elif value in (ResourceRoles.INTERMEDIATE_DATA, "i"):
            return ResourceRoles.INTERMEDIATE_DATA
        else:
            self.fail("Invalid resource role. Must be one of: %s" % ", ".join(ROLE_DATA_CHOICES))


DATA_ROLE_PARAM = DataRoleParamType()


@click.group()
@click.option(
    "-b",
    "--batch",
    default=False,
    is_flag=True,
    help="Run in batch mode, never ask for user inputs.",
)
@click.option(
    "--verbose",
    default=False,
    is_flag=True,
    help="Print extra debugging information and ask for confirmation before running actions.",
)
@click.pass_context
def cli(ctx, batch, verbose):
    ctx.obj = Namespace()
    ctx.obj.batch = batch
    ctx.obj.verbose = verbose
    global VERBOSE_MODE
    VERBOSE_MODE = verbose


class HostParamType(click.ParamType):
    name = "HOSTNAME"
    # name = "Must start with a letter or number and only contain letters, numbers, '.', '_', or '-'"

    def convert(self, value, param, ctx):
        if not HOSTNAME_RE.match(value):
            self.fail(
                "Must start with a letter or number and only contain letters, numbers, '.', '_', or '-'"
            )
        return value


HOST_PARAM = HostParamType()


class SnapshotParamType(click.ParamType):
    name = "SNAPSHOT"
    # name = "Must start with a letter or number and only contain letters, numbers, '.', '_', or '-'"

    def convert(self, value, param, ctx):
        # use the hostname re as that's a good approximation of the rules for tags
        if not HOSTNAME_RE.match(value):
            self.fail(
                "Snapshot tag or hash: must start with a letter or number and only contain letters, numbers, '.', '_', or '-'"
            )
        return value


SNAPSHOT_PARAM = SnapshotParamType()


class ResourceParamType(click.ParamType):
    name = "resources"

    def convert(self, value: Union[str, Sequence], param: Optional[click.Parameter], ctx):
        parsed = []
        if isinstance(value, str):
            rl = value.lower().split(",")  # type: Sequence[str]
        elif isinstance(value, Sequence):
            rl = value
        else:
            self.fail(
                "Invalid resource role list '%s', must be a string or a sequence" % str(value)
            )
        for r in rl:
            if r == "all":
                return [r for r in RESOURCE_ROLE_CHOICES]
            elif r in (ResourceRoles.SOURCE_DATA_SET, "s"):
                parsed.append(ResourceRoles.SOURCE_DATA_SET)
            elif r in (ResourceRoles.INTERMEDIATE_DATA, "i"):
                parsed.append(ResourceRoles.INTERMEDIATE_DATA)
            elif r in (ResourceRoles.CODE, "c"):
                parsed.append(ResourceRoles.CODE)
            elif r in (ResourceRoles.RESULTS, "r"):
                parsed.append(ResourceRoles.RESULTS)
            else:
                self.fail(
                    "Invalid resource role. Must be one of: %s, all"
                    % ", ".join(RESOURCE_ROLE_CHOICES)
                )
        return parsed


RESOURCE_PARAM = ResourceParamType()


@click.command()
@click.option(
    "--hostname",
    type=HOST_PARAM,
    default=None,
    help="Hostname to identify this machine in snapshot directory paths, "
    + "defaults to the result of the 'hostname' command.",
)
@click.option(
    "--create-resources",
    default=[],
    type=RESOURCE_PARAM,
    help="Initialize the workspace with subdirectories for the specified resource roles. Choices are 'all' or any comma-separated combination of %s."
    % ", ".join(RESOURCE_ROLE_CHOICES),
)
@click.option(
    "--scratch-directory",
    default=None,
    type=LocalPathType(allow_multiple_levels_of_missing_dirs=True),
    help="Local scratch directory (defaults to WORKSPACE_DIR/scratch)",
)
@click.option(
    "--git-fat-remote",
    type=str,
    help="Initialize the workspace with the git-fat large file extension "
    + "and use the specified URL for the remote datastore",
)
@click.option(
    "--git-fat-user", type=str, default=None, help="Username for git fat remote (if not root)"
)
@click.option(
    "--git-fat-port", type=int, default=None, help="Port number for git-fat remote (defaults to 22)"
)
@click.option(
    "--git-fat-attributes",
    type=str,
    default=None,
    help="Comma-separated list of file patterns to manage under git-fat."
    + " For example --git-fat-attributes='*.gz,*.zip'. If you do not specify"
    + " here, you can always add the .gitattributes file later.",
)
@click.option(
    "--git-lfs-attributes",
    type=str,
    default=None,
    help="Comma-separated list of file patterns to manage under git-lfs."
    + " For example --git-lfs-attributes='*.gz,*.zip'. If you do not specify"
    + " here, you can always add the .gitattributes file later.",
)
@click.argument("name", default=CURR_DIRNAME)
@click.pass_context
def init(
    ctx,
    hostname,
    name,
    create_resources,
    scratch_directory,
    git_fat_remote,
    git_fat_user,
    git_fat_port,
    git_fat_attributes,
    git_lfs_attributes,
):
    """Initialize a new workspace"""
    if hostname is None:
        if not ctx.obj.batch:
            hostname = click.prompt(
                "What name do you want to use for this machine in snapshot directory paths?",
                default=DEFAULT_HOSTNAME,
                type=HOST_PARAM,
            )
        else:
            hostname = DEFAULT_HOSTNAME
    if (
        (git_fat_user is not None) or (git_fat_port is not None) or (git_fat_attributes is not None)
    ) and (git_fat_remote is None):
        raise click.BadOptionUsage(
            message="If you specify --git-fat-user, --git-fat-port-, or --git-fat-attributes, "
            + "you also need to specify --git-fat-remote",
            option_name="--git-fat-remote",
        )
    init_command(
        name,
        hostname,
        create_resources,
        scratch_directory,
        git_fat_remote,
        git_fat_user,
        git_fat_port,
        git_fat_attributes,
        git_lfs_attributes,
        batch=ctx.obj.batch,
        verbose=ctx.obj.verbose,
    )


cli.add_command(init)


# The add command has subcommands for each resource type.
# This should be dynamically extensible, but we will hard
# code things for now.
@click.group()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.pass_context
def add(ctx, workspace_dir):
    """Add a data collection to the workspace as a resource. 
       Possible types of resources are ``git``, ``local-files``, or ``rclone``; these are subcommands of add."""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )

    ns.workspace_dir = workspace_dir


cli.add_command(add)


@click.command(name="local-files")
@click.option("--role", type=ROLE_PARAM)
@click.option("--name", type=str, default=None, help="Short name for this resource")
@click.option(
    "--compute-hash",
    is_flag=True,
    default=False,
    help="Compute hashes for all files. If this option is not set, we use a lightweight comparison of file sizes only.",
)
@click.option(
    "--export",
    "-e",
    is_flag=True,
    default=False,
    help="On snapshots, export lineage data for import into other workspaces",
)
@click.option(
    "--imported",
    is_flag=True,
    default=False,
    help="This resource was exported from another workspace. Import the lineage data. "
    + "An imported resource implies --read-only and --role=source-data.",
)
@click.argument("path", type=DIRECTORY_PARAM)
@click.pass_context
def local_files(ctx, role, name, compute_hash: bool, export: bool, imported: bool, path: str):
    """Add a local file directory (not managed by git) to the workspace. Subcommand of ``add``"""
    ns = ctx.obj
    if role is None:
        if imported:
            role = ResourceRoles.SOURCE_DATA_SET
        elif ns.batch:
            raise BatchModeError("--role")
        else:
            role = click.prompt(
                "Please enter a role for this resource, one of [s]ource-data, [i]ntermediate-data, [c]ode, or [r]esults",
                type=ROLE_PARAM,
            )
    path = abspath(expanduser(path))
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    if export and role in (ResourceRoles.SOURCE_DATA_SET, ResourceRoles.CODE):
        raise click.BadOptionUsage(
            message="Cannot export a source data or code resource", option_name="export"
        )
    if export and imported:
        raise click.BadOptionUsage(
            message="Cannot specify both --export and --imported", option_name="imported"
        )
    if imported and role != ResourceRoles.SOURCE_DATA_SET:
        raise click.BadOptionUsage(
            message="--imported only for source-data roles", option_name="imported"
        )
    add_command("file", role, name, workspace, path, compute_hash, export, imported)


add.add_command(local_files)


@click.command()
@click.option("--role", type=ROLE_PARAM)
@click.option("--name", type=str, default=None, help="Short name for this resource")
@click.option("--config", type=str, default=None, help="Configuration file for rclone")
@click.option("--compute-hash", is_flag=True, default=False, help="Compute hashes for all files")
@click.option(
    "--export",
    "-e",
    is_flag=True,
    default=False,
    help="On snapshots, export lineage data for import into other workspaces",
)
@click.option(
    "--imported",
    is_flag=True,
    default=False,
    help="This resource was exported from another workspace. Import the lineage data. "
    + "An imported resource implies --read-only and --role=source-data.",
)
@click.option(
    "--master",
    type=click.Choice(("none", "remote", "local"), case_sensitive=False),
    default="none",
    help="Determines which system is the master. If 'remote', then pulls will be done, but not pushes. "
    + "If 'local', then pushes will be done, but not pulls. If 'none' (the default), no action will be taken "
    + "for pushes and pulls (you need to synchronize manually using rclone). When first adding the resource "
    + " or cloning to a new machine, if the local directory does not exist, and 'remote' or 'none' were specified, "
    + " the contents of the remote will copied down to the local directory.",
)
@click.option(
    "--sync-mode",
    type=click.Choice(("copy", "sync"), case_sensitive=False),
    default="copy",
    help="When copying between local and master, which rclone command to use. If you specify 'copy', files are "
    + "added or overwritten without deleting any files present at the target. If you specify 'sync', files at "
    + "the target are removed if they are not present at the source. The default is 'copy'. If master is 'none', "
    + "this option has no effect.",
)
@click.option(
    "--size-only",
    is_flag=True,
    default=False,
    help="If specified, use only the file size (rather than also modification time and checksum) to "
    + "determine if a file has been changed. If your resource has a lot of files and access to the remote "
    + "is over a WAN, you probably want to set this. Otherwise, syncs/copies can be VERY slow.",
)
@click.argument("remote", type=str)
@click.argument(
    "local_path", type=str
)  # Currently, dest is required. Later: make dest optional and use the same path as remote?
@click.pass_context
def rclone(
    ctx,
    role,
    name,
    config: str,
    compute_hash: bool,
    export: bool,
    imported: bool,
    master: str,
    sync_mode: str,
    size_only: bool,
    remote: str,
    local_path: str,
):
    """Add an rclone-d repository as a resource to the workspace. Subcommand of ``add``.
    This is designed for uni-directional synchronization between a remote and a local_path.
    The remote has the form remote_name:remote_path, where remote_name is an entry in your
    rclone config file.
    """
    ns = ctx.obj
    if role is None:
        if imported:
            role = ResourceRoles.SOURCE_DATA_SET
        elif ns.batch:
            raise BatchModeError("--role")
        else:
            role = click.prompt(
                "Please enter a role for this resource, one of [s]ource-data, [i]ntermediate-data, [c]ode, or [r]esults",
                type=ROLE_PARAM,
            )
    rclone_re = r".*:.*"
    if re.match(rclone_re, remote) == None:
        raise click.BadOptionUsage(
            message="Source in rclone should be specified as remotename:filepath",
            option_name="source",
        )
    if export and role in (ResourceRoles.SOURCE_DATA_SET, ResourceRoles.CODE):
        raise click.BadOptionUsage(
            message="Cannot export a source data or code resource", option_name="export"
        )
    if export and imported:
        raise click.BadOptionUsage(
            message="Cannot specify both --export and --imported", option_name="imported"
        )
    if imported and role != ResourceRoles.SOURCE_DATA_SET:
        raise click.BadOptionUsage(
            message="--imported only for source-data roles", option_name="imported"
        )
    local_path = abspath(expanduser(local_path))
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    add_command(
        "rclone",
        role,
        name,
        workspace,
        remote,
        local_path,
        config,
        compute_hash,
        export,
        imported,
        master,
        sync_mode,
        size_only,
    )


add.add_command(rclone)


@click.command()
@click.option("--role", type=ROLE_PARAM)
@click.option("--name", type=str, default=None, help="Short name for this resource")
@click.option(
    "--branch", type=str, default=None,
    help="Branch of the repo to use. If not specified, defaults to the current branch."
)
@click.option(
    "--read-only",
    "-r",
    is_flag=True,
    default=False,
    help="If specified, treat the origin repository as read-only and never push to it.",
)
@click.option(
    "--export",
    "-e",
    is_flag=True,
    default=False,
    help="On snapshots, export lineage data for import into other workspaces",
)
@click.option(
    "--imported",
    is_flag=True,
    default=False,
    help="This resource was exported from another workspace. Import the lineage data. "
    + "An imported resource implies --read-only and --role=source-data.",
)
@click.argument("path", type=str)
@click.pass_context
def git(ctx, role, name, branch, read_only, export, imported, path):
    """Add a local git repository as a resource. Subcommand of ``add``"""
    ns = ctx.obj
    if role is None:
        if imported:
            role = ResourceRoles.SOURCE_DATA_SET
        elif ns.batch:
            raise BatchModeError("--role")
        else:
            role = click.prompt(
                "Please enter a role for this resource, one of [s]ource-data, [i]ntermediate-data, [c]ode, or [r]esults",
                type=ROLE_PARAM,
            )
    if export and role in (ResourceRoles.SOURCE_DATA_SET, ResourceRoles.CODE):
        raise click.BadOptionUsage(
            message="Cannot export a source data or code resource", option_name="export"
        )
    if export and imported:
        raise click.BadOptionUsage(
            message="Cannot specify both --export and --imported", option_name="imported"
        )
    if imported and role != ResourceRoles.SOURCE_DATA_SET:
        raise click.BadOptionUsage(
            message="--imported only for source-data roles", option_name="imported"
        )
    if imported:
        read_only = True
    if path.startswith("git@") or path.startswith("https://"):
        raise click.BadOptionUsage(
            message="It looks like you tried to specify a git URL (%s)." % path
            + " Currently, git resources only accept a local path."
            + " Try cloning your repository and then pasing the local path to that repository.",
            option_name="path",
        )

    path = abspath(expanduser(path))
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    add_command("git", role, name, workspace, path, branch, read_only, export, imported)


add.add_command(git)


@click.command(name="api-resource")
@click.option("--role", type=DATA_ROLE_PARAM)
@click.option("--name", type=str, default=None, help="Short name for this resource")
@click.pass_context
def api_resource(ctx, role, name):
    """Resource to represent data obtained via an API. Use this when there is
    no file-based representation of your data that can be versioned and captured
    more directly. Subcommand of ``add``"""
    ns = ctx.obj
    if role is None:
        if ns.batch:
            raise BatchModeError("--role")
        else:
            role = click.prompt(
                "Please enter a role for this resource, either [s]ource-data or [i]ntermediate-data",
                type=DATA_ROLE_PARAM,
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    add_command("api-resource", role, name, workspace)


add.add_command(api_resource)



@click.command(name="s3")
@click.option("--role", type=ROLE_PARAM)
@click.option("--name", type=str, default=None, help="Short name for this resource")
@click.argument("bucket_name", type=str)
@click.pass_context
def s3(ctx, role, name, bucket_name: str):
    """Add a S3 resource to the workspace. Subcommand of ``add``"""
    ns = ctx.obj
    if role is None:
        if ns.batch:
            raise BatchModeError("--role")
        else:
            role = click.prompt(
                "Please enter a role for this resource, one of [s]ource-data, [i]ntermediate-data, [c]ode, or [r]esults",
                type=ROLE_PARAM,
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    add_command("s3", role, name, workspace, bucket_name)


add.add_command(s3)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option("--message", "-m", type=str, default="", help="Message describing the snapshot")
@click.argument("tag", type=HOST_PARAM, default=None, required=False)
@click.pass_context
def snapshot(ctx, workspace_dir, message, tag):
    """Take a snapshot of the current workspace's state"""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    snapshot_command(workspace, tag, message)


cli.add_command(snapshot)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option(
    "--no-include-resources",
    is_flag=True,
    default=False,
    help="If specified, do NOT include deleting an snapshot-specific content from resources.",
)
@click.argument("tag_or_hash", type=HOST_PARAM, default=None, required=True)
@click.pass_context
def delete_snapshot(ctx, workspace_dir: str, no_include_resources: bool, tag_or_hash: str):
    """Delete the specified snapshot. This includes the metadata and lineage
    data for the snapshot. Unless --no-include-resources is specified, this
    also deletes any results data saved for the snapshot (under the
    snapshots subdirectory of a results resource)."""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    delete_snapshot_command(workspace, tag_or_hash, no_include_resources)


cli.add_command(delete_snapshot)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option(
    "--only",
    type=str,
    default=None,
    metavar="RESOURCE1[,RESOURCE2,...]",
    help="Comma-separated list of resource names that you wish to revert to the specified snapshot. The rest will be left as-is.",
)
@click.option(
    "--leave",
    type=str,
    default=None,
    metavar="RESOURCE1[,RESOURCE2,...]",
    help="Comma-separated list of resource names that you wish to leave in their current state. The rest will be restored to the specified snapshot.",
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    help="If specified, error out if unable to restore any of the requested resources "
    + "(due to lack of a restore hash or removing the resource from workspace).",
)
@click.argument("tag_or_hash", type=str, default=None, required=True)
@click.pass_context
def restore(
    ctx,
    workspace_dir: str,
    only: Optional[str],
    leave: Optional[str],
    strict: bool,
    tag_or_hash: str,
):
    """Restore the workspace to a prior state"""
    ns = ctx.obj
    if (only is not None) and (leave is not None):
        raise click.BadOptionUsage(option_name="--only", message="Please specify either --only or --leave, but not both")  # type: ignore
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    restore_command(
        workspace,
        tag_or_hash,
        only=only.split(",") if only else None,
        leave=leave.split(",") if leave else None,
        strict=strict,
    )


cli.add_command(restore)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option(
    "--skip",
    type=str,
    default=None,
    metavar="RESOURCE1[,RESOURCE2,...]",
    help="Comma-separated list of resource names that you wish to skip when pushing. The rest will be pushed to their remote origins, if applicable.",
)
@click.argument("remote-repository", type=str, default=None, required=True)
@click.pass_context
def publish(ctx, workspace_dir, skip: str, remote_repository):
    """Add a remote Git repository as the origin for the workspace and
    do the initial push of the workspace and any other resources.
    """
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    publish_command(workspace, remote_repository)
    push_command(workspace, only=None, skip=skip.split(",") if skip else None, only_workspace=False)


cli.add_command(publish)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option(
    "--only",
    type=str,
    default=None,
    metavar="RESOURCE1[,RESOURCE2,...]",
    help="Comma-separated list of resource names that you wish to push to the origin, if applicable. The rest will be skipped.",
)
@click.option(
    "--skip",
    type=str,
    default=None,
    metavar="RESOURCE1[,RESOURCE2,...]",
    help="Comma-separated list of resource names that you wish to skip when pushing. The rest will be pushed to their remote origins, if applicable.",
)
@click.option(
    "--only-workspace",
    is_flag=True,
    default=False,
    help="Only push the workspace's metadata, skipping the individual resources",
)
@click.pass_context
def push(ctx, workspace_dir: str, only: Optional[str], skip: Optional[str], only_workspace: bool):
    """Push the state of the workspace and its resources to their origins."""
    ns = ctx.obj
    option_cnt = (
        (1 if only is not None else 0)
        + (1 if skip is not None else 0)
        + (1 if only_workspace else 0)
    )
    if option_cnt > 1:
        raise click.BadOptionUsage(message="Please specify at most one of --only, --skip, or --only-workspace", option_name="--only")  # type: ignore
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    push_command(
        workspace,
        only=only.split(",") if only else None,
        skip=skip.split(",") if skip else None,
        only_workspace=only_workspace,
    )


cli.add_command(push)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option(
    "--only",
    type=str,
    default=None,
    metavar="RESOURCE1[,RESOURCE2,...]",
    help="Comma-separated list of resource names that you wish to pull from the origin, if applicable. The rest will be skipped.",
)
@click.option(
    "--skip",
    type=str,
    default=None,
    metavar="RESOURCE1[,RESOURCE2,...]",
    help="Comma-separated list of resource names that you wish to skip when pulling. The rest will be pulled from their remote origins, if applicable.",
)
@click.option(
    "--only-workspace",
    is_flag=True,
    default=False,
    help="Only pull the workspace's metadata, skipping the individual resources",
)
@click.pass_context
def pull(ctx, workspace_dir: str, only: Optional[str], skip: Optional[str], only_workspace: bool):
    """Pull the latest state of the workspace and its resources from their origins."""
    ns = ctx.obj
    option_cnt = (
        (1 if only is not None else 0)
        + (1 if skip is not None else 0)
        + (1 if only_workspace else 0)
    )
    if option_cnt > 1:
        raise click.BadOptionUsage(message="Please specify at most one of --only, --skip, or --only-workspace", option_name="--only")  # type: ignore
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    pull_command(
        workspace,
        only=only.split(",") if only else None,
        skip=skip.split(",") if skip else None,
        only_workspace=only_workspace,
    )


cli.add_command(pull)


@click.command()
@click.option(
    "--hostname",
    type=HOST_PARAM,
    default=None,
    help="Hostname to identify this machine in snapshot directory paths, "
    + "defaults to "
    + DEFAULT_HOSTNAME,
)
@click.argument("repository", type=str, default=None, required=True)
@click.argument("directory", type=str, default=None, required=False)
@click.pass_context
def clone(ctx, hostname, repository, directory):
    """Clone the specified data workspace."""
    ns = ctx.obj
    if hostname is None:
        if not ns.batch:
            hostname = click.prompt(
                "What name do you want to use for this machine in snapshot directory paths?",
                default=DEFAULT_HOSTNAME,
                type=HOST_PARAM,
            )
        else:
            hostname = DEFAULT_HOSTNAME
    clone_command(
        "dataworkspaces.backends.git", hostname, ns.batch, ns.verbose, repository, directory
    )


cli.add_command(clone)


# The report command has subcommands for specific reports on the workspace
@click.group()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.pass_context
def report(ctx, workspace_dir):
    """Report generation commands"""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )

    ns.workspace_dir = workspace_dir


cli.add_command(report)


@click.command(name="status")
@click.pass_context
def report_status(ctx):
    """Show the status of resources in this workspace. Subcommand of ``report``."""
    ns = ctx.obj
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    report_status_command(workspace)


report.add_command(report_status)


@click.command(name="history")
@click.option(
    "--limit",
    type=int,
    default=None,
    help="Number of previous snapshots to show (most recent first)",
)
@click.pass_context
def report_history(ctx, limit):
    """Show the history of snapshots. Subcommand of ``report``."""
    ns = ctx.obj
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    report_history_command(workspace, limit)


report.add_command(report_history)


@click.command(name="lineage")
@click.option(
    "--snapshot",
    type=SNAPSHOT_PARAM,
    default=None,
    help="Optional tag or hash for a snapshot. Otherwise, shows the current status.",
)
@click.pass_context
def report_lineage(ctx, snapshot=None):
    """Show a lineage table for either the current workspace or a specific snapshot.
    Subcommand of ``report``."""
    ns = ctx.obj
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    report_lineage_command(workspace, snapshot)


report.add_command(report_lineage)


@click.command(name="results")
@click.option(
    "--snapshot",
    type=SNAPSHOT_PARAM,
    default=None,
    help="Optional tag or hash for a snapshot. Otherwise, shows the current status.",
)
@click.option(
    "--resource",
    type=str,
    default=None,
    help="Optional resource name. Otherwise, will look for first resource with a results file.",
)
@click.pass_context
def report_results(ctx, snapshot=None, resource=None):
    """Show the contents of a results file.  Subcommand of ``report``."""
    ns = ctx.obj
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    report_results_command(workspace, snapshot, resource)


report.add_command(report_results)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option("--history", is_flag=True, default=False, help="Show previous snapshots")
@click.option(
    "--limit",
    type=int,
    default=None,
    help="Number of previous snapshots to show (most recent first)",
)
@click.pass_context
def status(ctx, workspace_dir, history, limit):
    """NOTE: this command is DEPRECATED. Please use ``dws report status`` and ``dws report history`` instead."""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    status_command(workspace, history, limit)


cli.add_command(status)


@click.command()
def version():
    """Print the version of Data Workspaces and exit.
    """
    click.echo("Data Workspaces version %s" % __version__)
    sys.exit(0)


cli.add_command(version)

# Disable run command for now, until we better understand how it interacts with the
# Lineage API. TODO: re-enable with the proper integration.
# @click.command()
# @click.option('--workspace-dir', type=WORKSPACE_PARAM, default=DWS_PATHDIR)
# @click.option('--step-name', default=None,
#               help="Name of step to associate with this command (defaults to base name of command without file extension)")
# @click.option('--cwd', default=CURR_DIR,
#               type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True),
#               help="Directory to use as current working directory when running command (defaults to %s)"%CURR_DIR)
# @click.argument('command', type=str, required=True)
# @click.argument('command_arguments', type=str, nargs=-1)
# @click.pass_context
# def run(ctx, workspace_dir, step_name, cwd, command, command_arguments):
#     """Run a data pipeline step from the command line, recording lineage information."""
#     ns = ctx.obj
#     if workspace_dir is None:
#         if ns.batch:
#             raise BatchModeError("--workspace-dir")
#         else:
#             workspace_dir = click.prompt("Please enter the workspace root dir",
#                                          type=WORKSPACE_PARAM)
#     run_command(workspace_dir, step_name, cwd, command, command_arguments, ns.batch, ns.verbose)

# cli.add_command(run)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.argument("snapshot_or_tag1", metavar="SNAPSHOT_OR_TAG1", type=str)
@click.argument("snapshot_or_tag2", metavar="SNAPSHOT_OR_TAG2", type=str)
@click.pass_context
def diff(ctx, workspace_dir, snapshot_or_tag1, snapshot_or_tag2):
    """List differences between two snapshots"""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    diff_command(workspace, snapshot_or_tag1, snapshot_or_tag2)


cli.add_command(diff)

# from .commands.show import show_command
# @click.command()
# @click.option('--workspace-dir', type=WORKSPACE_PARAM, default=DWS_PATHDIR)
# @click.pass_context
# def show(ctx, workspace_dir, limit):
#     """Show the current state of the workspace"""
#     ns = ctx.obj
#     show_command(workspace_dir, limit, ns.batch, ns.verbose)
#
# cli.add_command(show)


# The lineage command has subcommands for specific tasks
# with the linage.
@click.group()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.pass_context
def lineage(ctx, workspace_dir):
    """Lineage-related commands"""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )

    ns.workspace_dir = workspace_dir


cli.add_command(lineage)


@click.command(name="graph")
@click.option(
    "--resource",
    type=str,
    default=None,
    help="name of the resource to graph the lineage for (default to the first results resource)",
)
@click.option(
    "--snapshot",
    type=SNAPSHOT_PARAM,
    default=None,
    help="Snapshot hash or tag to use for lineage. If not specified, use current lineage.",
)
@click.option(
    "--format",
    default="html",
    type=click.Choice(["html", "dot"]),
    help="Format of the output graph (defaults to html)",
)
@click.option("--width", type=int, default=1024, help="Width of graph in pixels (defaults to 1024)")
@click.option("--height", type=int, default=800, help="Height of graph in pixels (defaults to 800)")
@click.argument(
    "output_file",
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
)
@click.pass_context
def graph(
    ctx,
    resource: Optional[str],
    snapshot: Optional[str],
    format: str,
    width: int,
    height: int,
    output_file: str,
):
    """Graph the lineage of a resource, writing the graph to an HTML file. Subcommand of ``lineage``"""
    ns = ctx.obj
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    lineage_graph_command(workspace, output_file, resource, snapshot, format, width, height)


lineage.add_command(graph)


# The deploy command has subcommands for specific tasks related to deployment
@click.group()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.pass_context
def deploy(ctx, workspace_dir):
    """Lineage-related commands"""
    ns = ctx.obj
    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    ns.workspace_dir = workspace_dir


cli.add_command(deploy)


@click.command(name="build")
@click.option(
    "--image-name",
    type=str,
    default=None,
    help="Name of docker image, defaults to name of workspace",
)
@click.option(
    "--force-rebuild",
    "-f",
    is_flag=True,
    default=False,
    help="If specified, always rebuild image (force deletes the image from docker)",
)
@click.option(
    "--git-user-email",
    type=str,
    default=None,
    help="Email address used by git inside the container. Defaults to value of user.email for this workspace.",
)
@click.option(
    "--git-user-name",
    type=str,
    default=None,
    help="Username used by git inside the container. Defualts to value of user.name for this workspace.",
)
@click.pass_context
def deploy_build(
    ctx,
    image_name: Optional[str],
    force_rebuild: bool,
    git_user_email: Optional[str],
    git_user_name: Optional[str],
):
    """Build a docker image containing this workspace. This command is EXERIMENTAL
    and subject to change."""
    ns = ctx.obj
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    deploy_build_command(workspace, image_name, force_rebuild, git_user_email, git_user_name)


deploy.add_command(deploy_build)


@click.command(name="run")
@click.option(
    "--image-name",
    type=str,
    default=None,
    help="Name of docker image, defaults to name of workspace",
)
@click.option(
    "--no-mount-ssh-keys",
    is_flag=True,
    default=False,
    help="If specified, do not mount the host's ~/.ssh directory into the container. This directory is need for git authentication.",
)
@click.pass_context
def deploy_run(ctx, image_name: Optional[str], no_mount_ssh_keys: bool):
    """Build a docker image containing this workspace. This command is EXERIMENTAL
    and subject to change."""
    ns = ctx.obj
    workspace = find_and_load_workspace(ns.batch, ns.verbose, ns.workspace_dir)
    deploy_run_command(workspace, image_name, no_mount_ssh_keys)


deploy.add_command(deploy_run)


@click.command()
@click.option("--workspace-dir", type=WORKSPACE_PARAM, default=DWS_PATHDIR)
@click.option(
    "--resource",
    type=str,
    default=None,
    help="If specified, get/set parameters for the specified resource, rather than the workspace.",
)
@click.argument("param_name", metavar="[PARAMETER_NAME]", default=None, required=False)
@click.argument("param_value", metavar="[PARAMETER_VALUE]", default=None, required=False)
@click.pass_context
def config(ctx, workspace_dir, resource, param_name, param_value):
    """Get or set configuration parameters. Local parameters are only for this
    copy of the workspace, while global parameters are stored centrally and
    affect all copies.

    If neither PARAMETER_NAME nor PARAMETER_VALUE are specified, this command
    prints a table of all parameters and their information (scope, value, default or not,
    and help text). If just PARAMETER_NAME is specified, it prints the specified parameter's
    information. Finally, if both the parameter name and value are specified, the parameter
    is set to the specified value."""
    ns = ctx.obj

    if workspace_dir is None:
        if ns.batch:
            raise BatchModeError("--workspace-dir")
        else:
            workspace_dir = click.prompt(
                "Please enter the workspace root dir", type=WORKSPACE_PARAM
            )
    workspace = find_and_load_workspace(ns.batch, ns.verbose, workspace_dir)
    config_command(workspace, param_name, param_value, resource)


cli.add_command(config)


if __name__ == "__main__":
    cli()
    sys.exit(0)



#!/usr/bin/env python3
"""
Python 3 wrapper over git-fat (which is python2-only)
"""
import os
import sys
from os.path import dirname, abspath, expanduser, isfile, join, exists, curdir
import subprocess
import click
from typing import Optional, Dict, Any

from dataworkspaces.errors import ConfigurationError, InternalError

THIS_FILES_DIRPATH=dirname(abspath(expanduser(__file__)))

STANDARD_EXE_SEARCH_LOCATIONS=['/usr/bin', '/usr/local/bin',
                               abspath(expanduser('~/bin'))]


def _is_python2(exe_path, verbose=False):
    try:
        cp = subprocess.run([exe_path, '--version'], cwd=THIS_FILES_DIRPATH,
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                            encoding='utf-8')
        cp.check_returncode()
        output = cp.stdout
    except Exception as e:
        if verbose:
            click.echo("Got an error when trying to run python at %s: %s"%
                       (exe_path, e))
        return False
    if output.startswith('Python 2.7') or output.startswith('Python 2.6'):
        return True
    else:
        return False

NOT_FOUND_ERRORMSG=\
"""Did not find a Python 2 executable, which is required to run git-fat.
Looked for executables in the directories %s.
Please make sure that you have Python 2.7 or Python 2.6 installed and
the conrtaining directory referenced on your PATH environment variable.
"""

def find_python2_exe():
    dirpaths = os.environ['PATH'].split(os.pathsep)
    for dirpath in STANDARD_EXE_SEARCH_LOCATIONS:
        if dirpath not in dirpaths:
            dirpaths.append(dirpath)

    for dirpath in dirpaths:
        for exe_name in ['python2.7', 'python2', 'python']:
            exe_path = join(dirpath, exe_name)
            if isfile(exe_path) and os.access(exe_path, os.X_OK):
                if _is_python2(exe_path):
                    return exe_path
    raise ConfigurationError(NOT_FOUND_ERRORMSG%', '.join(dirpaths))

def run_git_fat(python2_exe, args, cwd=curdir, verbose=False):
    fat_script = join(THIS_FILES_DIRPATH, 'git-fat')
    if not exists(fat_script):
        raise InternalError("Missing %s" % fat_script)
    cmd = [python2_exe, fat_script]+args
    if verbose:
        click.echo("%s from %s" % (' '.join(cmd), cwd))
        env = os.environ.copy() # type: Optional[Dict[str,Any]]
        assert env is not None
        env['GIT_FAT_VERBOSE'] = "1"
    else:
        env = None
    cp = subprocess.run(cmd, cwd=cwd, env=env)
    try:
        cp.check_returncode()
    except Exception as e:
        raise InternalError("git-fat execution with args '%s' failed." %
                            ' '.join(args)) from e

def main(args=sys.argv[1:]):
    """Just run git-fat with the command line arguments.
    """
    python2=find_python2_exe()
    run_git_fat(python2, args)


if __name__=='__main__':
    sys.exit(main())



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Resource for git repositories
"""
import subprocess
import os
from os.path import (
    realpath,
    basename,
    isdir,
    isfile,
    join,
    dirname,
    exists,
    abspath,
    expanduser,
    commonpath,
    isabs,
)
import shutil
import stat
import click
import json
from typing import Set, Pattern, Union, Optional, Tuple, cast, List

from dataworkspaces.errors import ConfigurationError, InternalError, PathError
from dataworkspaces.utils.subprocess_utils import call_subprocess, call_subprocess_for_rc
from dataworkspaces.utils.git_utils import (
    is_git_dirty,
    is_file_tracked_by_git,
    get_local_head_hash,
    get_branch_info,
    commit_changes_in_repo,
    checkout_and_apply_commit,
    GIT_EXE_PATH,
    is_git_repo,
    commit_changes_in_repo_subdir,
    checkout_subdir_and_apply_commit,
    get_subdirectory_hash,
    is_pull_needed_from_remote,
    git_remove_subtree,
    git_remove_file,
    git_commit,
    git_add,
    is_git_staging_dirty,
    switch_git_branch,
    switch_git_branch_if_needed,
)
from dataworkspaces.utils.git_fat_utils import (
    is_a_git_fat_repo,
    has_git_fat_been_initialized,
    validate_git_fat_in_path,
    validate_git_fat_in_path_if_needed,
)
from dataworkspaces.utils.git_lfs_utils import ensure_git_lfs_configured_if_needed
from dataworkspaces.workspace import (
    Resource,
    ResourceFactory,
    ResourceRoles,
    RESOURCE_ROLE_PURPOSES,
    LocalStateResourceMixin,
    FileResourceMixin,
    SnapshotResourceMixin,
    SnapshotWorkspaceMixin,
    JSONDict,
    JSONList,
    Workspace,
)
import dataworkspaces.backends.git as git_backend
from dataworkspaces.utils.file_utils import (
    LocalPathType,
    does_subpath_exist,
    get_subpath_from_absolute,
)
from dataworkspaces.utils.param_utils import BoolType, AbspathType, StringType, RelpathType

from dataworkspaces.utils.snapshot_utils import (
    move_current_files_local_fs,
    copy_current_files_local_fs,
)


def git_move_and_add(srcabspath, destabspath, git_root, verbose):
    """
    Move a file that might or might not be tracked by git to
    a new location (snapshot directory), set it to read-only and make sure
    that it is now tracked by git.
    """
    assert srcabspath.startswith(git_root)
    assert destabspath.startswith(git_root)
    srcrelpath = srcabspath[len(git_root) + 1 :]
    destrelpath = destabspath[len(git_root) + 1 :]
    if is_file_tracked_by_git(srcrelpath, git_root, verbose):
        call_subprocess(
            [GIT_EXE_PATH, "mv", srcrelpath, destrelpath], cwd=git_root, verbose=verbose
        )
    else:
        # file is not tracked by git yet, just move and then add to git
        os.rename(join(git_root, srcrelpath), join(git_root, destrelpath))
    # either way, we change the permissions and then do an add at the end
    mode = os.stat(destabspath)[stat.ST_MODE]
    os.chmod(destabspath, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
    git_add(git_root, [destrelpath], verbose)


def git_copy_and_add(srcabspath, destabspath, git_root, verbose):
    """
    This is for an exported results resource.
    Copy a file that might or might not be tracked by git to
    a new location (snapshot directory), set it to read-only and make sure
    that it is now tracked by git. We add it to the original location if it is not
    already in the repo.
    """
    assert srcabspath.startswith(git_root)
    assert destabspath.startswith(git_root)
    srcrelpath = srcabspath[len(git_root) + 1 :]
    destrelpath = destabspath[len(git_root) + 1 :]
    shutil.copyfile(srcabspath, destabspath)
    add_files = [destrelpath]
    if not is_file_tracked_by_git(srcrelpath, git_root, verbose):
        add_files.append(srcrelpath)
    mode = os.stat(destabspath)[stat.ST_MODE]
    os.chmod(destabspath, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
    git_add(git_root, add_files, verbose)


class GitResourceBase(Resource, LocalStateResourceMixin, FileResourceMixin, SnapshotResourceMixin):
    def __init__(
        self,
        resource_type: str,
        name: str,
        role: str,
        workspace: Workspace,
        local_path: str,
        export: bool,
        repo_dir: str,
    ):
        super().__init__(resource_type, name, role, workspace)

        # define and validate the parameters
        self.param_defs.define(
            "export",
            default_value=False,
            optional=True,
            help="True if metadata for export should be added each snapshot",
            is_global=True,
            ptype=BoolType(),
        )
        self.export = self.param_defs.get("export", export)  # type: bool
        self.param_defs.define(
            "local_path",
            default_value=None,
            optional=False,
            help="Always points to the root of this resource "
            + "(not necessarily the root of the repo)",
            is_global=False,
            ptype=AbspathType(),
        )
        self.local_path = self.param_defs.get("local_path", local_path)  # type: str

        self.repo_dir = repo_dir  # The root of the repo.

    def get_local_path_if_any(self):
        return self.local_path

    def validate_subpath_exists(self, subpath: str) -> None:
        super().validate_subpath_exists(subpath)

    def delete_snapshot(
        self, workspace_snapshot_hash: str, resource_restore_hash: str, relative_path: str
    ) -> None:
        snapshot_dir_path = join(self.local_path, relative_path)
        if isdir(snapshot_dir_path):
            if self.workspace.verbose:
                print(
                    "Deleting snapshot directory %s from resource %s" % (relative_path, self.name)
                )
            subpath_relative_to_repo = get_subpath_from_absolute(self.repo_dir, snapshot_dir_path)
            assert subpath_relative_to_repo is not None
            git_remove_subtree(
                self.repo_dir, subpath_relative_to_repo, verbose=self.workspace.verbose
            )
            git_commit(
                self.repo_dir, "Deleted %s" % snapshot_dir_path, verbose=self.workspace.verbose
            )

    def does_subpath_exist(
        self, subpath: str, must_be_file: bool = False, must_be_directory: bool = False
    ) -> bool:
        return does_subpath_exist(self.local_path, subpath, must_be_file, must_be_directory)

    def upload_file(self, src_local_path: str, rel_dest_path: str) -> None:
        """Copy a local file to the specified path in the
        resource. This may be a local copy or an upload, depending
        on the resource implmentation
        """
        abs_dest_path = join(self.local_path, rel_dest_path)
        parent_dir = dirname(abs_dest_path)
        if not exists(src_local_path):
            raise ConfigurationError("Source file %s does not exist" % src_local_path)
        if not exists(parent_dir):
            os.makedirs(parent_dir)
        shutil.copyfile(src_local_path, abs_dest_path)
        rel_to_repo_path = get_subpath_from_absolute(self.repo_dir, abs_dest_path)
        assert rel_to_repo_path is not None
        call_subprocess(
            [GIT_EXE_PATH, "add", rel_to_repo_path],
            cwd=self.repo_dir,
            verbose=self.workspace.verbose,
        )
        if is_git_staging_dirty(self.repo_dir, rel_to_repo_path):
            call_subprocess(
                [GIT_EXE_PATH, "commit", "-m", "Added %s" % rel_to_repo_path],
                cwd=self.repo_dir,
                verbose=self.workspace.verbose,
            )
        if self.workspace.verbose:
            click.echo("%s: Copied file to %s" % (self.name, rel_dest_path))

    def read_results_file(self, subpath: str) -> JSONDict:
        """Read and parse json results data from the specified path
        in the resource. If the path does not exist or is not a file
        throw a ConfigurationError.
        """
        path = join(self.local_path, subpath)
        if not isfile(path):
            raise ConfigurationError(
                "subpath %s does not exist or is not a file in resource %s" % (subpath, self.name)
            )
        with open(path, "r") as f:
            try:
                return json.load(f)
            except Exception as e:
                raise ConfigurationError(
                    "Parse error when reading %s in resource %s" % (subpath, self.name)
                ) from e

    def delete_file(self, rel_path: str) -> None:
        """Delete a file from the resource. If the resource is read-only or
        otherwise does not support modifications, should throw a NotSupportedError.
        """
        if is_file_tracked_by_git(rel_path, self.local_path, self.workspace.verbose):
            git_remove_file(self.local_path, rel_path, verbose=self.workspace.verbose)
        else:
            os.remove(join(self.local_path, rel_path))

    def ls(self, rel_path:str) -> List[str]:
        if rel_path!='':
            path = join(self.local_path, rel_path)
        else:
            path = self.local_path
        if isdir(path):
            files = os.listdir(path)
            if rel_path!='':
                return [join(rel_path, file) for file in files]
            else:
                return files
        elif exists(path):
            return [rel_path]
        else:
            raise PathError(f"Path {rel_path} does not exist in resource {self.name}")

    def open(self, rel_path:str, mode:str):
        if rel_path=='':
            raise PathError("Need to specify a file to open, empty string is not valid here")
        path = join(self.local_path, rel_path)
        return open(path, mode)


def get_workspace_dir(workspace: Workspace) -> str:
    workspace_local_path = workspace.get_workspace_local_path_if_any()
    return workspace_local_path if workspace_local_path is not None else abspath(expanduser("~"))


class GitRepoResource(GitResourceBase):
    def __init__(
        self,
        name: str,
        role: str,
        workspace: Workspace,
        remote_origin_url: str,
        relative_local_path: Optional[str],
        local_path: str,
        branch: str,
        read_only: bool,
        export: bool,
        imported: bool,
    ):
        super().__init__(
            "git", name, role, workspace, local_path, export=export, repo_dir=local_path
        )

        # handle parameters
        self.param_defs.define(
            "remote_origin_url",
            default_value=None,
            optional=False,
            help="URL of the remote git repo",
            is_global=True,
            ptype=StringType(),
        )
        self.remote_origin_url = self.param_defs.get(
            "remote_origin_url", remote_origin_url
        )  # type: str
        self.param_defs.define(
            "relative_local_path",
            default_value=None,
            optional=True,
            help="Local path of repo relative to the workspace",
            is_global=True,
            ptype=RelpathType(),
        )
        self.relative_local_path = self.param_defs.get(
            "relative_local_path", relative_local_path
        )  # type: Optional[str]
        self.param_defs.define(
            "branch",
            default_value=None,
            optional=False,
            help="Git branch to use",
            is_global=True,
            ptype=StringType(),
        )
        self.branch = self.param_defs.get("branch", branch)  # type: str
        self.param_defs.define(
            "read_only",
            default_value=False,
            optional=True,
            help="If True, than no pushes are done for this repo",
            is_global=True,
            ptype=BoolType(),
        )
        self.read_only = self.param_defs.get("read_only", read_only)  # type: bool
        self.param_defs.define(
            "imported",
            default_value=False,
            optional=True,
            help="If True, then this resource has lineage imported from another workspace",
            is_global=True,
            ptype=BoolType(),
        )
        self.imported = self.param_defs.get("imported", imported)  # type: bool

    def get_local_params(self):
        use_relative = (
            True
            if self.relative_local_path
            and join(get_workspace_dir(self.workspace), self.relative_local_path) == self.local_path
            else False
        )
        params = super().get_local_params()
        params["local_path"] = self.relative_local_path if use_relative else self.local_path
        return params

    def results_move_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ):
        switch_git_branch_if_needed(self.local_path, self.branch, self.workspace.verbose)
        validate_git_fat_in_path_if_needed(self.local_path)
        moved_files = move_current_files_local_fs(
            self.name,
            self.local_path,
            rel_dest_root,
            exclude_files,
            exclude_dirs_re,
            move_fn=lambda src, dest: git_move_and_add(
                src, dest, self.local_path, self.workspace.verbose
            ),
            verbose=self.workspace.verbose,
        )
        # If there were no files in the results dir, then we do not
        # create a subdirectory for this snapshot
        if len(moved_files) > 0:
            call_subprocess(
                [GIT_EXE_PATH, "commit", "-m", "Move current results to %s" % rel_dest_root],
                cwd=self.local_path,
                verbose=self.workspace.verbose,
            )

    def results_copy_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ):
        switch_git_branch_if_needed(self.local_path, self.branch, self.workspace.verbose)
        validate_git_fat_in_path_if_needed(self.local_path)
        copied_files = copy_current_files_local_fs(
            self.name,
            self.local_path,
            rel_dest_root,
            exclude_files,
            exclude_dirs_re,
            copy_fn=lambda src, dest: git_copy_and_add(
                src, dest, self.local_path, self.workspace.verbose
            ),
            verbose=self.workspace.verbose,
        )
        # If there were no files in the results dir, then we do not
        # create a subdirectory for this snapshot
        if len(copied_files) > 0:
            call_subprocess(
                [GIT_EXE_PATH, "commit", "-m", "Copy current results to %s" % rel_dest_root],
                cwd=self.local_path,
                verbose=self.workspace.verbose,
            )

    def add_results_file(self, data: Union[JSONDict, JSONList], rel_dest_path: str) -> None:
        """Save JSON results data to the specified path in the resource.
        """
        # TODO: We are abusing add results file for other things, see workspace.py
        # assert self.role == ResourceRoles.RESULTS
        switch_git_branch_if_needed(self.local_path, self.branch, self.workspace.verbose)
        abs_dest_path = join(self.local_path, rel_dest_path)
        parent_dir = dirname(abs_dest_path)
        if not exists(parent_dir):
            os.makedirs(parent_dir)
        with open(abs_dest_path, "w") as f:
            json.dump(data, f, indent=2)
        call_subprocess(
            [GIT_EXE_PATH, "add", rel_dest_path],
            cwd=self.local_path,
            verbose=self.workspace.verbose,
        )
        call_subprocess(
            [GIT_EXE_PATH, "commit", "-m", "Added %s" % rel_dest_path],
            cwd=self.local_path,
            verbose=self.workspace.verbose,
        )

    def snapshot_precheck(self):
        validate_git_fat_in_path_if_needed(self.local_path)

    def snapshot(self):
        # Todo: handle tags
        commit_changes_in_repo(
            self.local_path, "autocommit ahead of snapshot", verbose=self.workspace.verbose
        )
        switch_git_branch_if_needed(self.local_path, self.branch, self.workspace.verbose)
        hashval = get_local_head_hash(self.local_path, self.workspace.verbose)
        return (hashval, hashval)

    def restore_precheck(self, hashval):
        rc = call_subprocess_for_rc(
            [GIT_EXE_PATH, "cat-file", "-e", hashval + "^{commit}"],
            cwd=self.local_path,
            verbose=self.workspace.verbose,
        )
        if rc != 0:
            raise ConfigurationError("No commit found with hash '%s' in %s" % (hashval, str(self)))
        if is_a_git_fat_repo(self.local_path):
            import dataworkspaces.third_party.git_fat as git_fat

            self.python2_exe = git_fat.find_python2_exe()
            self.uses_git_fat = True
            validate_git_fat_in_path()
        else:
            self.uses_git_fat = False

    def restore(self, hashval):
        commit_changes_in_repo(
            self.local_path, "auto-commit ahead of restore", verbose=self.workspace.verbose
        )
        switch_git_branch_if_needed(self.local_path, self.branch, self.workspace.verbose)
        checkout_and_apply_commit(self.local_path, hashval, verbose=self.workspace.verbose)
        if self.uses_git_fat:
            # since the restored repo might have different git-fat managed files, we run
            # a pull to get them.
            import dataworkspaces.third_party.git_fat as git_fat

            git_fat.run_git_fat(
                self.python2_exe, ["pull"], cwd=self.local_path, verbose=self.workspace.verbose
            )

    def push_precheck(self):
        if self.read_only:
            return
        if is_git_dirty(self.local_path):
            raise ConfigurationError(
                "Git repo at %s has uncommitted changes. Please commit your changes before pushing."
                % self.local_path
            )
        if is_pull_needed_from_remote(self.local_path, self.branch, self.workspace.verbose):
            raise ConfigurationError(
                "Resource '%s' requires a pull from the remote origin before pushing." % self.name
            )
        if is_a_git_fat_repo(self.local_path):
            import dataworkspaces.third_party.git_fat as git_fat

            self.python2_exe = git_fat.find_python2_exe()
            self.uses_git_fat = True
        else:
            self.uses_git_fat = False

    def push(self):
        """Push to remote origin, if any"""
        if self.read_only:
            click.echo("Skipping push of resource %s, as it is read-only" % self.name)
            return
        switch_git_branch_if_needed(self.local_path, self.branch, self.workspace.verbose)
        call_subprocess(
            [GIT_EXE_PATH, "push", "origin", self.branch],
            cwd=self.local_path,
            verbose=self.workspace.verbose,
        )
        if self.uses_git_fat:
            import dataworkspaces.third_party.git_fat as git_fat

            git_fat.run_git_fat(
                self.python2_exe, ["push"], cwd=self.local_path, verbose=self.workspace.verbose
            )

    def pull_precheck(self):
        if is_git_dirty(self.local_path):
            raise ConfigurationError(
                "Git repo at %s has uncommitted changes. Please commit your changes before pulling."
                % self.local_path
            )
        if is_a_git_fat_repo(self.local_path):
            import dataworkspaces.third_party.git_fat as git_fat

            self.python2_exe = git_fat.find_python2_exe()
            self.uses_git_fat = True
        else:
            self.uses_git_fat = False

    def pull(self):
        """Pull from remote origin, if any"""
        switch_git_branch_if_needed(self.local_path, self.branch, self.workspace.verbose)
        call_subprocess(
            [GIT_EXE_PATH, "pull", "origin", self.branch],
            cwd=self.local_path,
            verbose=self.workspace.verbose,
        )
        if self.uses_git_fat:
            import dataworkspaces.third_party.git_fat as git_fat

            git_fat.run_git_fat(
                self.python2_exe, ["pull"], cwd=self.local_path, verbose=self.workspace.verbose
            )

    def __str__(self):
        return "Git repository %s in role '%s'" % (self.local_path, self.role)


def get_remote_origin(local_path, verbose=False):
    args = [GIT_EXE_PATH, "config", "--get", "remote.origin.url"]
    if verbose:
        click.echo(" ".join(args) + " [run in %s]" % local_path)
    cp = subprocess.run(
        args, cwd=local_path, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    if cp.returncode != 0:
        click.echo("Remote origin not found for git repo at %s" % local_path)
        return None
    return cp.stdout.rstrip()



class GitLocalPathType(LocalPathType):
    def __init__(self, remote_url: str, verbose: bool):
        super().__init__()
        self.remote_url = remote_url
        self.verbose = verbose

        def convert(self, value, param, ctx):
            rv = super().convert(value, param, ctx)
            if isdir(rv):
                if not isdir(join(rv, ".git")):
                    self.fail(
                        '%s "%s" exists, but is not a git repository' % (self.path_type, rv),
                        param,
                        ctx,
                    )
                remote = get_remote_origin(rv, verbose=self.verbose)
                if remote != self.remote_url:
                    self.fail(
                        '%s "%s" is a git repo with remote origin "%s", but dataworkspace has remote "%s"'
                        % (self.path_type, rv, self.remote_url, remote),
                        param,
                        ctx,
                    )
            return rv


class GitRepoFactory(ResourceFactory):
    def from_command_line(
        self, role, name, workspace, local_path, branch, read_only, export, imported
    ):
        """Instantiate a resource object from the add command's
        arguments"""
        workspace.validate_local_path_for_resource(name, local_path)
        lpr = realpath(local_path)
        wspath = (
            realpath(workspace.get_workspace_local_path_if_any())
            if workspace.get_workspace_local_path_if_any() is not None
            else None
        )
        if not is_git_repo(local_path):
            if (
                isinstance(workspace, git_backend.Workspace)
                and wspath is not None
                and lpr.startswith(wspath)
            ):
                dws_branch = workspace.get_dws_git_branch()
                if (branch is not None ) and (branch != dws_branch):
                    raise ConfigurationError(
                        f"Branch for subdirectory must match brach of primary dws git repository ({dws_branch})"
                    )
                elif read_only:
                    raise ConfigurationError(
                        "The --read-only parameter is only valid for separate git repositories, not subdirectories."
                    )
                elif imported:
                    raise ConfigurationError(
                        "The --imported parameter is only valid for separate git repositories, not subdirectories."
                    )
                return GitRepoSubdirFactory().from_command_line(
                    role, name, workspace, local_path, export=export
                )
            else:
                raise ConfigurationError(local_path + " is not a git repository")
        # The local path is a git repo. Double-check that it isn't already part
        # of the workspace's repo. If it is, you will get an error when cloning.
        if (
            isinstance(workspace, git_backend.Workspace)
            and wspath is not None
            and lpr.startswith(wspath)
            and is_file_tracked_by_git(
                local_path, workspace.get_workspace_local_path_if_any(), verbose=workspace.verbose
            )
        ):
            raise ConfigurationError(
                "%s is a git repository, but also part of the parent workspace's repo"
                % (local_path)
            )
        validate_git_fat_in_path_if_needed(local_path)
        ensure_git_lfs_configured_if_needed(local_path, verbose=workspace.verbose)
        remote_origin = get_remote_origin(local_path, verbose=workspace.verbose)
        if remote_origin is None:
            raise ConfigurationError(
                "Remote origin is missing for git repository at %s. " % local_path
                + "Please configure before adding resource."
            )
        (current, others) = get_branch_info(local_path, workspace.verbose)
        if branch is None:
            branch = current # we default to the current branch
        if branch != current and branch not in others:
            raise ConfigurationError(
                "Requested branch '%s' is not available for git repository at %s"
                % (branch, local_path)
            )
        if is_git_dirty(local_path) and branch != current:
            click.echo(
                "WARNING: Git repo is currently on branch %s and branch %s was requested. However, the current branch has uncommitted changes. Will skip changing the branch after adding the repo to workspace."
                % (current, branch)
            )
        else:
            switch_git_branch(local_path, branch, workspace.verbose)
        if imported:
            lineage_path = join(local_path, "lineage.json")
            if not exists(lineage_path):
                raise ConfigurationError(
                    "--imported was specified, but missing exported lineage file %s" % lineage_path
                )
            if (
                not isinstance(workspace, SnapshotWorkspaceMixin)
                or not workspace.supports_lineage()
            ):
                raise ConfigurationError(
                    "--imported was specified, but this workspace does not support lineage"
                )
            with open(lineage_path, "r") as f:
                lineage_data = json.load(f)
            if lineage_data["resource_name"] != name:
                raise ConfigurationError(
                    "Resource name in imported lineage '%s' does not match '%s'"
                    % (lineage_data["resource_name"], name)
                )
            cast(SnapshotWorkspaceMixin, workspace).get_lineage_store().import_lineage_file(
                name, lineage_data["lineages"]
            )
        workspace_dir = get_workspace_dir(workspace)
        relative_local_path = None  # type: Optional[str]
        if commonpath([workspace_dir, local_path]) == workspace_dir:
            relative_local_path = get_subpath_from_absolute(workspace_dir, local_path)
        return GitRepoResource(
            name,
            role,
            workspace,
            remote_origin,
            relative_local_path,
            local_path,
            branch,
            read_only,
            export,
            imported,
        )

    def from_json(self, params, local_params, workspace):
        """Instantiate a resource object from the parsed resources.json file"""
        assert params["resource_type"] == "git"
        local_path = local_params["local_path"]
        if not isabs(local_path):
            local_path = join(get_workspace_dir(workspace), local_path)
        return GitRepoResource(
            params["name"],
            params["role"],
            workspace,
            params["remote_origin_url"],
            params.get("relative_local_path", None),
            local_path,
            params["branch"],
            params.get("read_only", False),
            params.get("export", False),
            params.get("imported", False),
        )

    def has_local_state(self):
        return True

    def clone(self, params, workspace):
        assert params["resource_type"] == "git"
        rname = params["name"]
        remote_origin_url = params["remote_origin_url"]
        relative_local_path = params.get("relative_local_path", None)
        workspace_dir = get_workspace_dir(workspace)
        default_local_path = (
            join(workspace_dir, relative_local_path)
            if relative_local_path
            else join(workspace_dir, rname)
        )
        branch = params["branch"]
        read_only = params.get("read_only", False)
        if not workspace.batch:
            # ask the user for a local path
            local_path = click.prompt(
                "Git resource '%s' is being added to your workspace. Where do you want to clone it?"
                % rname,
                default=default_local_path,
                type=GitLocalPathType(remote_origin_url, workspace.verbose),
            )
        else:
            if isdir(default_local_path):
                if not isdir(join(default_local_path, ".git")):
                    raise ConfigurationError(
                        "Unable to add resource '%s' as default local path '%s' exists but is not a git repository."
                        % (rname, default_local_path)
                    )
                remote = get_remote_origin(default_local_path, workspace.verbose)
                if remote != remote_origin_url:
                    raise ConfigurationError(
                        "Unable to add resource '%s' as remote origin in local path '%s' is %s, but data workspace has '%s'"
                        % (rname, default_local_path, remote, remote_origin_url)
                    )
            local_path = default_local_path
        parent = dirname(local_path)
        if not exists(local_path):
            # cloning a fresh repository
            cmd = [GIT_EXE_PATH, "clone", remote_origin_url, basename(local_path)]
            call_subprocess(cmd, parent, workspace.verbose)
        else:
            # the repo already exists locally, and we've alerady verified that then
            # remote is correct
            cmd = [GIT_EXE_PATH, "pull", "origin", branch]
            call_subprocess(cmd, local_path, workspace.verbose)
        switch_git_branch_if_needed(local_path, branch, workspace.verbose, ok_if_not_present=True)
        ensure_git_lfs_configured_if_needed(local_path, verbose=workspace.verbose)
        if is_a_git_fat_repo(local_path) and not has_git_fat_been_initialized(local_path):
            import dataworkspaces.third_party.git_fat as git_fat

            python2_exe = git_fat.find_python2_exe()
            git_fat.run_git_fat(python2_exe, ["init"], cwd=local_path, verbose=workspace.verbose)
            git_fat.run_git_fat(python2_exe, ["pull"], cwd=local_path, verbose=workspace.verbose)

        return GitRepoResource(
            rname,
            params["role"],
            workspace,
            remote_origin_url,
            relative_local_path,
            local_path,
            branch,
            read_only,
            params.get("export", False),
            params.get("imported", False),
        )

    def suggest_name(self, workspace, role, local_path, branch, read_only, export, imported):
        return basename(local_path)


def _get_workspace_dir_for_git_backend(workspace):
    """This is used by the git-subdirectory resources, which only work with the
    git backend for the workspace...
    """
    if not isinstance(workspace, git_backend.Workspace):
        raise ConfigurationError(
            "Git subdirectory resources are only supported with the Git workspace backend."
        )
    workspace_dir = workspace.get_workspace_local_path_if_any()
    assert workspace_dir is not None
    return workspace_dir


class GitRepoResultsSubdirResource(GitResourceBase):
    """Resource for a subdirectory of the workspace for when it is
     in the results role.
    """

    def __init__(self, name: str, workspace: Workspace, relative_path: str, export: bool):
        # only valid when workspace has git backend
        workspace_dir = _get_workspace_dir_for_git_backend(workspace)
        super().__init__(
            "git-subdirectory",
            name,
            ResourceRoles.RESULTS,
            workspace,
            join(workspace_dir, relative_path),
            export=export,
            repo_dir=workspace_dir,
        )
        self.workspace_dir = workspace_dir
        self.param_defs.define(
            "relative_path",
            default_value=None,
            optional=False,
            help="Path of resource's directory relative to the workspace root",
            is_global=True,
            ptype=RelpathType(),
        )
        self.relative_path = self.param_defs.get("relative_path", relative_path)  # type: str

    def results_move_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ):
        validate_git_fat_in_path_if_needed(self.workspace_dir)
        moved_files = move_current_files_local_fs(
            self.name,
            self.local_path,
            rel_dest_root,
            exclude_files,
            exclude_dirs_re,
            move_fn=lambda src, dest: git_move_and_add(
                src, dest, self.local_path, self.workspace.verbose
            ),
            verbose=self.workspace.verbose,
        )
        # If there were no files in the results dir, then we do not
        # create a subdirectory for this snapshot
        if len(moved_files) > 0:
            call_subprocess(
                [
                    GIT_EXE_PATH,
                    "commit",
                    "--only",
                    self.relative_path,
                    "-m",
                    "Move current results to %s" % rel_dest_root,
                ],
                cwd=self.workspace_dir,
                verbose=self.workspace.verbose,
            )

    def results_copy_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ):
        validate_git_fat_in_path_if_needed(self.workspace_dir)
        copied_files = copy_current_files_local_fs(
            self.name,
            self.local_path,
            rel_dest_root,
            exclude_files,
            exclude_dirs_re,
            copy_fn=lambda src, dest: git_copy_and_add(
                src, dest, self.local_path, self.workspace.verbose
            ),
            verbose=self.workspace.verbose,
        )
        # If there were no files in the results dir, then we do not
        # create a subdirectory for this snapshot
        if len(copied_files) > 0:
            call_subprocess(
                [
                    GIT_EXE_PATH,
                    "commit",
                    "--only",
                    self.relative_path,
                    "-m",
                    "Copy current results to %s" % rel_dest_root,
                ],
                cwd=self.workspace_dir,
                verbose=self.workspace.verbose,
            )

    def snapshot_precheck(self):
        validate_git_fat_in_path_if_needed(self.workspace_dir)

    def snapshot(self):
        # The subdirectory hash is used for comparison and the head
        # hash used for restoring
        return (
            get_subdirectory_hash(
                self.workspace_dir, self.relative_path, verbose=self.workspace.verbose
            ),
            get_local_head_hash(self.workspace_dir, verbose=self.workspace.verbose),
        )

    def restore_precheck(self, hashval):
        raise ConfigurationError(
            "Git subdirectory resource '%s' should not be included in restore set" % self.name
        )

    def restore(self, hashval):
        raise InternalError(
            "Should never call restore on a git subdirectory resource (%s)" % self.name
        )

    def get_branch(self):
        """The branch will be the same as our parent"""
        assert isinstance(self.workspace, git_backend.Workspace)
        return self.workspace.get_dws_git_branch()

    def add_results_file(self, data: Union[JSONDict, JSONList], rel_dest_path: str) -> None:
        """Save JSON results data to the specified path in the resource.
        """
        abs_dest_path = join(self.local_path, rel_dest_path)
        parent_dir = dirname(abs_dest_path)
        if not exists(parent_dir):
            os.makedirs(parent_dir)
        with open(abs_dest_path, "w") as f:
            json.dump(data, f, indent=2)
        rel_to_repo_path = join(self.relative_path, rel_dest_path)
        call_subprocess(
            [GIT_EXE_PATH, "add", rel_to_repo_path],
            cwd=self.workspace_dir,
            verbose=self.workspace.verbose,
        )
        call_subprocess(
            [GIT_EXE_PATH, "commit", "-m", "Added %s" % rel_to_repo_path],
            cwd=self.workspace_dir,
            verbose=self.workspace.verbose,
        )

    def push_precheck(self):
        if not exists(self.local_path):
            raise ConfigurationError(
                "Missing directory %s for resource %s" % (self.local_path, self.name)
            )
        # if the backend is a git repo, we skip the check here, as we check at the repo level
        # and ask the user whether they want to proceed anyway.
        if is_git_dirty(self.workspace_dir) and not isinstance(
            self.workspace, git_backend.Workspace
        ):
            raise ConfigurationError(
                "Git repo at %s has uncommitted changes. Please commit your changes before pushing."
                % self.workspace_dir
            )
        if is_pull_needed_from_remote(self.workspace_dir, self.get_branch(), self.workspace.verbose):
            raise ConfigurationError(
                "Resource '%s' requires a pull from the remote origin before pushing." % self.name
            )

    def push(self):
        """Push to remote origin, if any"""
        pass  # push will happen at workspace level

    def pull_precheck(self):
        if is_git_dirty(self.local_path):
            raise ConfigurationError(
                "Git repo at %s has uncommitted changes. Please commit your changes before pulling."
                % self.workspace_dir
            )

    def pull(self):
        """Pull from remote origin, if any"""
        pass  # pull will happen at workspace level

    def __str__(self):
        return "Git repository subdirectory %s in role '%s'" % (self.relative_path, self.role)


class GitRepoSubdirResource(GitResourceBase):
    """Resource for a subdirectory of the workspace for when it is NOT
    in the results role.
    """

    def __init__(
        self, name: str, role: str, workspace: Workspace, relative_path: str, export: bool
    ):
        assert role != ResourceRoles.RESULTS
        workspace_dir = _get_workspace_dir_for_git_backend(workspace)
        super().__init__(
            "git-subdirectory",
            name,
            role,
            workspace,
            join(workspace_dir, relative_path),
            export=export,
            repo_dir=workspace_dir,
        )
        self.workspace_dir = workspace_dir
        self.param_defs.define(
            "relative_path",
            default_value=None,
            optional=False,
            help="Path of resource's directory relative to the workspace root",
            is_global=True,
            ptype=RelpathType(),
        )
        self.relative_path = self.param_defs.get("relative_path", relative_path)  # type: str

    def results_move_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ):
        raise InternalError(
            "results_move_current_files should not be called for %s" % self.__class__.__name__
        )

    def results_copy_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ):
        raise InternalError(
            "results_copy_current_files should not be called for %s" % self.__class__.__name__
        )

    def add_results_file(self, data: Union[JSONDict, JSONList], rel_dest_path: str) -> None:
        """Save JSON results data to the specified path in the resource.
        """
        abs_dest_path = join(self.local_path, rel_dest_path)
        parent_dir = dirname(abs_dest_path)
        if not exists(parent_dir):
            os.makedirs(parent_dir)
        with open(abs_dest_path, "w") as f:
            json.dump(data, f, indent=2)
        rel_to_repo_path = join(self.relative_path, rel_dest_path)
        call_subprocess(
            [GIT_EXE_PATH, "add", rel_to_repo_path],
            cwd=self.workspace_dir,
            verbose=self.workspace.verbose,
        )
        call_subprocess(
            [GIT_EXE_PATH, "commit", "-m", "Added %s" % rel_to_repo_path],
            cwd=self.workspace_dir,
            verbose=self.workspace.verbose,
        )

    def snapshot_precheck(self):
        validate_git_fat_in_path_if_needed(self.workspace_dir)

    def snapshot(self) -> Tuple[Optional[str], Optional[str]]:
        """Returns (cmopare_hash, restore_hash)
        """
        # Todo: handle tags
        commit_changes_in_repo_subdir(
            self.workspace_dir,
            self.relative_path,
            "autocommit ahead of snapshot",
            verbose=self.workspace.verbose,
        )
        return (
            get_subdirectory_hash(
                self.workspace_dir, self.relative_path, verbose=self.workspace.verbose
            ),
            get_local_head_hash(self.workspace_dir, verbose=self.workspace.verbose),
        )

    def restore_precheck(self, hashval):
        validate_git_fat_in_path_if_needed(self.workspace_dir)
        rc = call_subprocess_for_rc(
            [GIT_EXE_PATH, "cat-file", "-e", hashval + "^{commit}"],
            cwd=self.workspace_dir,
            verbose=self.workspace.verbose,
        )
        if rc != 0:
            raise ConfigurationError("No commit found with hash '%s' in %s" % (hashval, str(self)))

    def restore(self, hashval):
        commit_changes_in_repo_subdir(
            self.workspace_dir,
            self.relative_path,
            "auto-commit ahead of restore",
            verbose=self.workspace.verbose,
        )
        checkout_subdir_and_apply_commit(
            self.workspace_dir, self.relative_path, hashval, verbose=self.workspace.verbose
        )

    def get_branch(self):
        """The branch will be the same as our parent"""
        assert isinstance(self.workspace, git_backend.Workspace)
        return self.workspace.get_dws_git_branch()

    def push_precheck(self):
        if not exists(self.local_path):
            raise ConfigurationError(
                "Missing directory %s for resource %s" % (self.local_path, self.name)
            )
        # if the backend is a git repo, we skip the check here, as we check at the repo level
        # and ask the user whether they want to proceed anyway.
        if is_git_dirty(self.workspace_dir) and not isinstance(
            self.workspace, git_backend.Workspace
        ):
            raise ConfigurationError(
                "Git repo at %s has uncommitted changes. Please commit your changes before pushing."
                % self.local_path
            )
        if is_pull_needed_from_remote(self.local_path, self.get_branch(), self.workspace.verbose):
            raise ConfigurationError(
                "Resource '%s' requires a pull from the remote origin before pushing." % self.name
            )

    def push(self):
        """Push to remote origin, if any"""
        pass  # push will happen at workspace level
        # actions.call_subprocess([GIT_EXE_PATH, 'push', 'origin', 'master'],
        #                         cwd=self.local_path, verbose=self.verbose)

    def pull_precheck(self):
        if is_git_dirty(self.local_path):
            raise ConfigurationError(
                "Git repo at %s has uncommitted changes. Please commit your changes before pulling."
                % self.local_path
            )

    def pull(self):
        """Pull from remote origin, if any"""
        pass  # pull will happen at workspace level
        # actions.call_subprocess([GIT_EXE_PATH, 'pull', 'origin', 'master'],
        #                         cwd=self.local_path, verbose=self.verbose)

    def __str__(self):
        return "Git repository subdirectory %s in role '%s'" % (self.relative_path, self.role)


CONFIRM_SUBDIR_MSG = (
    "The subdirectory %s does not currently exist, but must be a part of the workspace's git repo in order"
    + " to be used as a resource. Do you want it to be created and added to git?"
)


def create_results_subdir(workspace_dir, full_path, relative_path, role, verbose):
    os.makedirs(full_path)
    with open(join(full_path, "README.txt"), "w") as f:
        f.write("This directory is for %s.\n" % RESOURCE_ROLE_PURPOSES[role])
        f.write(
            "This README file ensures the directory is added to the git repository, as git does not support empty directories.\n"
        )
    call_subprocess([GIT_EXE_PATH, "add", relative_path], cwd=workspace_dir, verbose=verbose)
    call_subprocess(
        [GIT_EXE_PATH, "commit", "-m", "Add %s to repo for storing results" % relative_path],
        cwd=workspace_dir,
        verbose=verbose,
    )
    click.echo("Added %s to git repository" % relative_path)


class GitRepoSubdirFactory(ResourceFactory):
    """This is a version of a git repo resource where we are just
    storing in a subdirectory of a repo rather than the full repo.
    This is currently only valid if we are storing as a subdir of the
    main data workspace repo.
    """

    def from_command_line(
        self, role, name, workspace, local_path, confirm_subdir_create=True, export=False
    ):
        """Instantiate a resource object from the add command's
        arguments"""
        if is_git_repo(local_path):
            raise InternalError(
                "Local path '%s'is a git repo, should not be using GitRepoSubdirFactory"
                % local_path
            )
        lpr = realpath(local_path)
        workspace_dir = _get_workspace_dir_for_git_backend(workspace)
        validate_git_fat_in_path_if_needed(workspace_dir)
        ensure_git_lfs_configured_if_needed(workspace_dir, verbose=workspace.verbose)
        wdr = realpath(workspace_dir)
        if not lpr.startswith(wdr):
            raise ConfigurationError(
                "Git subdirectories can only be used as resources when under the workspace repo."
            )
        relative_path = lpr[len(wdr) + 1 :]
        if not exists(local_path):
            if not confirm_subdir_create:
                create_results_subdir(
                    workspace_dir, local_path, relative_path, role, workspace.verbose
                )
            elif not workspace.batch:
                click.confirm(CONFIRM_SUBDIR_MSG % relative_path, abort=True)
                create_results_subdir(
                    workspace_dir, local_path, relative_path, role, workspace.verbose
                )
            else:
                raise ConfigurationError(
                    "Did not find '%s'. Cannot create a resource from a git subdirectory if the directory does not already exist."
                    % local_path
                )
        if role == ResourceRoles.RESULTS:
            return GitRepoResultsSubdirResource(name, workspace, relative_path, export)
        else:
            return GitRepoSubdirResource(name, role, workspace, relative_path, export)

    def from_json(self, params, local_params, workspace):
        """Instantiate a resource object from the parsed resources.json file"""
        assert params["resource_type"] == "git-subdirectory"
        if params["role"] == ResourceRoles.RESULTS:
            return GitRepoResultsSubdirResource(
                params["name"], workspace, params["relative_path"], params.get("export", False)
            )
        else:
            return GitRepoSubdirResource(
                params["name"],
                params["role"],
                workspace,
                params["relative_path"],
                params.get("export", False),
            )

    def clone(self, params, workspace):
        assert params["resource_type"] == "git-subdirectory"
        rname = params["name"]
        role = params["role"]
        relative_path = params["relative_path"]
        if not isinstance(workspace, git_backend.Workspace):
            raise ConfigurationError(
                "Git subdirectory resources are only supported with the Git workspace backend."
            )
        workspace_dir = workspace.get_workspace_local_path_if_any()
        assert workspace_dir is not None
        # this should be redundant, as we already intialized for the parent repo, but run just in case.
        ensure_git_lfs_configured_if_needed(workspace_dir, verbose=workspace.verbose)
        local_path = join(workspace_dir, relative_path)
        if not exists(local_path):
            # this subdirectory most have been created in the remote
            # resource. We can just wait for the "git pull" to populate the
            # the contents, but will create a placeholder so our checks pass.
            os.mkdir(local_path)
        if role == ResourceRoles.RESULTS:
            return GitRepoResultsSubdirResource(
                rname, workspace, relative_path, params.get("export", False)
            )
        else:
            return GitRepoSubdirResource(
                rname, role, workspace, relative_path, params.get("export", False)
            )

    def has_local_state(self) -> bool:
        return True

    def suggest_name(self, local_path, *args):
        return basename(local_path)
